from collections import defaultdict
from hedra.core.personas.batching.param_type import ParamType
from hedra.versioning.flags.types.unstable.flag import unstable
from numpy.random import normal
from statistics import mean
from typing import (
    Dict, 
    List, 
    Tuple, 
    Union, 
    Optional
)
from .base_algorithm import BaseAlgorithm


@unstable
class PointOptimizer(BaseAlgorithm):

    def __init__(
        self, 
        config: Dict[str, Union[List[Tuple[Union[int, float]]], int]],
        distribution_idx: Optional[int]=None
    ) -> None:
        super().__init__(
            config,
            distribution_idx=distribution_idx
        )

        self.params_history = defaultdict(list)
        self.next_params = {}
        self.error_history = defaultdict(list)
        self.stop_iteration = False
        self.error_batch_size_map = defaultdict(list)
        self.param_errors = defaultdict(lambda: defaultdict(list))
        self._noise_scale = 0.1
        self._target_value: int = None

    async def optimize(self, func):

        for param_name, boundary in self._boundaries_by_param_name.items():

            next_value = mean([boundary[0], boundary[1]])
            self.next_params[param_name] = next_value

            self.params_history[param_name].append(next_value)

        for param_name in self.optimize_params:
            for _ in range(self.max_iter):
                    error, target_value = await func(
                        list(self.next_params.values())
                    )

                    if error:
                        self._compute_next_params(
                            error,
                            target_value,
                            param_name
                        )

                    else:
                        boundary = self._boundaries_by_param_name.get(param_name)
                        next_value = mean([boundary[0], boundary[1]])
                        self.next_params[param_name] = next_value

                        self.params_history[param_name].append(next_value)

        final_params = {}
        minimized_error_mean = 0
        minimized_parameter = 0
        for param_name in self.optimize_params:

            param_errors = self.param_errors[param_name]
            param_error_sums = {}
            param_values_by_sum = defaultdict(list)

            for param_value, errors in param_errors.items():
                mean_errors = mean(errors)
                param_error_sums[param_value] = mean_errors
                param_values_by_sum[mean_errors].append(param_value)

            error_sums = list(param_error_sums.values())

            if len(error_sums) > 0:
                minimized_error_mean = min(error_sums)

            else:
                minimized_error_mean = 0
                minimized_parameter = self._target_value

            minimized_parameter = min(param_values_by_sum[minimized_error_mean])


            param = self.param_values.get(param_name, {})
            params_type = param.get('type')

            if params_type == ParamType.INTEGER:
                minimized_parameter = int(minimized_parameter)

            else:
                minimized_parameter = float(minimized_parameter)

            final_params[param_name] = {
                'minimized_distribution_value': minimized_parameter,
                'minimized_error_mean': minimized_error_mean
            }

        return final_params
            

    def _compute_next_params(
        self, 
        error: float, 
        target_value: Union[float, int],
        optimizing_param_name: str, 
    ):
        
        last_param_value = self.params_history[optimizing_param_name][-1]
        adjustment_ratio = 1
        absolute_error = abs(error)
        completed = error + target_value

        if self._target_value is None:
            self._target_value = target_value


        if error > 0:

            if target_value > absolute_error:
                adjustment_ratio = absolute_error/target_value

            else:
                adjustment_ratio = target_value/absolute_error

            noise_distribution = normal(
                scale=self._noise_scale,
                size=1
            )

            noise_value = noise_distribution[0]

            adjustment_ratio = abs(
                round(adjustment_ratio + noise_value, 2)
            )
            next_param_value = last_param_value * adjustment_ratio

        else:
            adjustment_ratio = absolute_error/target_value

            noise_distribution = normal(
                scale=self._noise_scale,
                size=1
            )

            noise_value = noise_distribution[0]            

            adjustment_ratio = abs(
                round(adjustment_ratio + noise_value, 2)
            )
            next_param_value = last_param_value * (1 + adjustment_ratio)
            
        minimum_boundary, maximum_boundary = self._boundaries_by_param_name[optimizing_param_name]

        if next_param_value < minimum_boundary:
            next_param_value = minimum_boundary

        elif next_param_value > maximum_boundary:
            next_param_value = maximum_boundary

        if next_param_value < 1:
            next_param_value = 1

        if completed > 0:
            self.next_params[optimizing_param_name] = next_param_value
            self.params_history[optimizing_param_name].append(next_param_value)
            self.error_history[optimizing_param_name].append(
                absolute_error
            )

            self.error_batch_size_map[absolute_error].append(last_param_value)
            self.param_errors[optimizing_param_name][last_param_value].append(absolute_error)





from enum import IntEnum


class SettingCodes(IntEnum):
    """
    All known HTTP/2 setting codes.

    .. versionadded:: 2.6.0
    """

     #: The byte that signals the SETTINGS_HEADER_TABLE_SIZE setting.
    HEADER_TABLE_SIZE = 0x01
    #: The byte that signals the SETTINGS_ENABLE_PUSH setting.
    ENABLE_PUSH = 0x02
    #: The byte that signals the SETTINGS_MAX_CONCURRENT_STREAMS setting.
    MAX_CONCURRENT_STREAMS = 0x03
    #: The byte that signals the SETTINGS_INITIAL_WINDOW_SIZE setting.
    INITIAL_WINDOW_SIZE = 0x04
    #: The byte that signals the SETTINGS_MAX_FRAME_SIZE setting.
    MAX_FRAME_SIZE = 0x05
    #: The byte that signals the SETTINGS_MAX_HEADER_LIST_SIZE setting.
    MAX_HEADER_LIST_SIZE = 0x06
    #: The byte that signals SETTINGS_ENABLE_CONNECT_PROTOCOL setting.
    ENABLE_CONNECT_PROTOCOL = 0x08



from typing import Any
from hedra.core.engines.types.websocket import WebsocketResult
from .http_processed_result import HTTPProcessedResult


class WebsocketProcessedResult(HTTPProcessedResult):

    def __init__(
        self, 
        stage: str, 
        result: WebsocketResult
    ) -> None:
        super(
            WebsocketProcessedResult,
            self
        ).__init__(
            stage,
            result
        )


import uuid
import time
import asyncio
import threading
import os
import signal
from typing import Any, Dict, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.personas.batching.param_type import ParamType
from hedra.core.personas import get_persona
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.logging import HedraLogger
from hedra.monitoring.base.exceptions import MonitorKilledError
from hedra.tools.data_structures import AsyncList
from .algorithms import get_algorithm
from .algorithms.types.base_algorithm import BaseAlgorithm


class Optimizer:

    def __init__(self, config: Dict[str, Any]) -> None:

        self.logger = HedraLogger()
        self.logger.initialize()

        self.optimizer_id = str(uuid.uuid4())
        self.thread_id = threading.current_thread().ident
        self.process_id = os.getpid()
        
        self.graph_name = config.get('graph_name')
        self.graph_id = config.get('graph_id')
        self.target_stage_name = config.get('stage_name')
        self.source_stage_name = config.get('source_stage_name')
        self.source_stage_id = config.get('source_stage_id')
        self.stage_config: Config = config.get('stage_config')
        self.stage_hooks = config.get('stage_hooks')
        self.base_batch_size = self.stage_config.batch_size

        self.metadata_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self.thread_id} - process:{self.process_id} - Stage: {self.source_stage_name}:{self.source_stage_id} - Optimizer: {self.optimizer_id} - '

        self.stage_name = config.get('stage_name')
        self.actions = AsyncList()

        self.algorithm_type = config.get('algorithm', 'shg')

        self._optimization_time_limit = config.get('time_limit', 60)

        self.algorithm: BaseAlgorithm = get_algorithm(
            self.algorithm_type,
            {
                **config,
                'stage_config': self.stage_config
            }
        )

        self._current_iter = 0
        self.optimized_results = {}
        self._max_aps = 0
        self._event_loop: asyncio.AbstractEventLoop = None

        self.current_params = {}
        self.start = 0
        self.elapsed = 0

    def optimize(self) -> Dict[str, Union[int, float]]:

        results = None

        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Starting optimization')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Time Limit - {self._optimization_time_limit}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Algorithm - {self.algorithm_type}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Batch Time - {self.algorithm.batch_time}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Max Iter - {self.algorithm.max_iter}')

        self.start = time.time()

        results = self.algorithm.optimize(self._run_optimize)

        optimized_params = {}
        for idx in range(len(results.x)):
            param_name = self.algorithm.param_names[idx]
            optimiazed_param_name = f'optimized_{param_name}'
            param = self.algorithm.param_values.get(param_name, {})
            param_type = param.get('type')

            if param_type == ParamType.INTEGER:
                optimized_params[optimiazed_param_name] = int(results.x[idx])

            else:
                optimized_params[optimiazed_param_name] = float(results.x[idx])

        self.total_optimization_time = time.time() - self.start

        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization took - {round(self.total_optimization_time, 2)} - seconds')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization - max actions per second - {self._max_aps}')

        for optimized_param_name, optimized_value in optimized_params.items():
            param_log_name = optimized_param_name.replace('_', ' ')
            self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization - {param_log_name} - {optimized_value}')

        self.optimized_results = {
            **optimized_params,
            'optimize_target_stage': self.target_stage_name,
            'optimization_iters': self.algorithm.max_iter,
            'optimization_iter_duation': self.algorithm.batch_time,
            'optimization_total_time': self.total_optimization_time,
            'optimization_max_aps': self._max_aps
        }

        self._event_loop.close()

        return self.optimized_results
    
    def _setup_persona(self, stage_config: Config) -> DefaultPersona:
        persona = get_persona(self.stage_config)
        persona.optimization_active = True
        persona.setup(self.stage_hooks, self.metadata_string)

        return persona
    
    def _handle_async_exception(self, loop, ctx) -> None:
        pass

    async def _optimize(self, xargs: List[Union[int, float]]) -> float:

        if self._current_iter < self.algorithm.max_iter and self.elapsed < self.algorithm.time_limit:

            persona = self._setup_persona(self.stage_config)

            for idx, param in enumerate(xargs):
                param_name = self.algorithm.param_names[idx]
                param = self.algorithm.param_values.get(param_name, {})
                param_type = param.get('type')

                if param_type == ParamType.INTEGER:
                    xargs[idx] = int(xargs[idx])

                else:
                    xargs[idx] = float(xargs[idx])

                param['value'] = xargs[idx]

                self.current_params[param_name] = xargs[idx]
                self.algorithm.current_params[param_name] = xargs[idx]
            
            persona = self.algorithm.update_params(persona)
            await persona.set_concurrency(persona.batch.size)

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter}')

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Size - {persona.batch.size}')
            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Interval - {persona.batch.interval}')
            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Gradient - {persona.batch.gradient}')

            completed_count = 0
            try:
                results = await persona.execute()
                completed_count = len([result for result in results if result.error is None])

            except RuntimeError as e:
                raise e
            
            except KeyboardInterrupt as e:
                raise e

            except Exception:
                pass

            elapsed = persona.end - persona.start

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - took - {round(elapsed, 2)} - seconds')
       
            if completed_count < 1:
                completed_count = 1

            if elapsed < 1:
                elapsed = float('inf')

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Inverted APS score- {elapsed/completed_count}')
            
            return elapsed/completed_count

        return self.base_batch_size

    def _run_optimize(self, xargs: List[Union[int, float]]) -> float:

        try:
            self._event_loop = asyncio.get_event_loop()
        except Exception:
            self._event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._event_loop)

            self._event_loop.set_exception_handler(self._handle_async_exception)

            def handle_loop_stop(signame):
                try:
                    self._event_loop.close()

                except KeyboardInterrupt:
                    raise RuntimeError()

                except BrokenPipeError:
                    raise RuntimeError()
                    
                except RuntimeError:
                    raise RuntimeError()
                
                except MonitorKilledError:
                    raise RuntimeError()

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):

                self._event_loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(signame)
                )

        inverse_actions_per_second = self._event_loop.run_until_complete(
            self._optimize(xargs)
        )

        self._current_iter += 1
        self.elapsed = time.time() - self.start
 
        return inverse_actions_per_second


import io
import struct
from typing import Dict
from .pack_string import pack_string

def pack_domain_name(
    name: bytes, 
    names: Dict[bytes, bytes], 
    offset: int=0
):

    parts = name.split('.')
    buf = io.BytesIO()

    while parts:

        subname = '.'.join(parts)
        u = names.get(subname)

        if u:
            buf.write(struct.pack('!H', 0xc000 + u))
            break

        else:
            names[subname] = buf.tell() + offset

        buf.write(
            pack_string(
                parts.pop(0)
            )
        )

    else:
        buf.write(b'\0')
        
    return buf.getvalue()



from __future__ import annotations
from typing import Union, Dict
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http.result import HTTPResult
from .action import HTTP3Action

try:
    from aioquic.h3.events import (
        HeadersReceived,
        DataReceived
    )

except ImportError:
    HeadersReceived = object
    DataReceived = object
    

class HTTP3Result(HTTPResult):

    __slots__ = (
        'action_id',
        'url',
        'ip_addr',
        'method',
        'path',
        'params',
        'query',
        'hostname',
        'headers',
        'headers_frame',
        'body',
        'response_code',
        '_version',
        '_reason',
        '_status'
    )

    def __init__(
            self, 
            action: HTTP3Action, 
            error: Exception = None
        ) -> None:
        super().__init__(
            action, 
            error
        )

        self.body = bytearray()
        self.headers_frame: HeadersReceived = None
        self.body: DataReceived = None
        self.type = RequestTypes.HTTP3

    @property
    def status(self) -> Union[int, None]:
        try:
            response_status = self.headers.get(b':status')
            if self._status is None and isinstance(response_status, (bytes, bytearray)):
                self._status = int(response_status)

        except Exception:
            pass

        return self._status
 
        


from .cassandra_graphql_action_schema import CassandraGraphQLActionSchema
from .cassandra_graphql_http2_action_schema import CassandraGraphQLHTTP2ActionSchema
from .cassandra_grpc_action_schema import CassandraGRPCActionSchema
from .cassandra_http_action_schema import CassandraHTTPActionSchema
from .cassandra_http2_action_schema import CassandraHTTP2ActionSchema
from .cassandra_http3_action_schema import CassandraHTTP3ActionSchema
from .cassandra_playwright_action_schema import CassandraPlaywrightActionSchema
from .cassandra_task_schema import CassandraTaskSchema
from .cassandra_udp_action_schema import CassandraUDPActionSchema
from .cassandra_websocket_action_schema import CassandraWebsocketActionSchema


import asyncio
import functools
import json
import os
import pathlib
import psutil
import signal
import re
import uuid
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.parsers.parser import Parser
from typing import (
    Dict, 
    List, 
    TextIO, 
    Union,
    Any,
    Coroutine
)
from .har_connector_config import HARConnectorConfig

try:

    from haralyzer import HarParser, HarEntry
    from haralyzer.http import Request

    has_connector=True

except ImportError:
    has_connector=False

    HarParser=object
    HarEntry=object
    Request=object


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop, 
    harfile: TextIO
): 
    try:
        harfile.close()
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass
    

class HARConnector:
    connector_type=ConnectorType.HAR

    def __init__(
        self, 
        config: HARConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.filepath = config.filepath
        self._name_pattern = re.compile('[^0-9a-zA-Z]+')
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.stage = stage
        self.parser_config = parser_config

        self.logger = HedraLogger()
        self.logger.initialize()

        self._parser: Union[HarParser, None] = None
        self.har_file: TextIO = None

        self._loop: asyncio.AbstractEventLoop = None

        self.parser = Parser()
        self._protocol_types_map: Dict[str, str] = {
            "http/3.0": "http3",
            "http/2.0": "http2",
            "http/1.1": "http"
        }

    async def connect(self):
        self._loop = asyncio._get_running_loop()

        if self.filepath[:2] == '~/':
            user_directory = pathlib.Path.home()
            self.filepath = os.path.join(
                user_directory,
                self.filepath[2:]
            )
        
        self.filepath = await self._loop.run_in_executor(
            None,
            functools.partial(
                os.path.abspath,
                self.filepath
            )
        )

        self.har_file = await self._loop.run_in_executor(
            None,
            functools.partial(
                open,
                self.filepath
            )
        )

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop,
                    self.har_file
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening from file - {self.filepath}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, NotImplementedError]:
        raise NotImplementedError('Execute stage summary loading is not available for the HAR Connector.')

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        
        self._parser = HarParser(
            har_data=json.loads(self.har_file)
        )
        
        actions: List[Dict[str, Any]] = []

        action_order = 0

        for page in self._parser.pages:
            page_entries: List[HarEntry] = page.entries

            for entry in page_entries:
                
                page_request: Request = entry.request

                action_data = page_request.text
                
                content_type: str = page_request.mimeType
                if content_type.lower() == 'application/json' and page_request.text:
                    action_data = json.loads(action_data)

                action_url: str = page_request.url
                action_method_stub: str= page_request.method
                action_method = action_method_stub.upper()
                action_headers: List[Dict[str, str]] = page_request.headers

                action_basename = '_'.join([
                    segment.capitalize() for segment in self._name_pattern.sub(
                        '_', 
                        action_url
                    ).split('_')
                ])

                action_fullname = f'{action_method.lower()}_{action_basename}'

                http_type: str = self._protocol_types_map.get(
                    page_request.httpVersion,
                    "http"
                )

                actions.append({
                    'engine': http_type,
                    'name': action_fullname,
                    'url': action_url,
                    'headers': action_headers,
                    'method': action_method,
                    'data': action_data,
                    'order': action_order
                })

                action_order += 1

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, NotImplementedError]:
        raise NotImplementedError('Results loading is not available for the HAR Connector.')
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, NotImplementedError]:
        raise NotImplementedError('Data loading is not available for the HAR Connector.')
    
    async def close(self):
        await self._loop.run_in_executor(
            self._executor,
            self.har_file.close
        )

        self._executor.shutdown(cancel_futures=True)



import zstandard
from base64 import b64encode
from hedra.distributed.middleware.base import (
    Middleware,
    MiddlewareType
)
from hedra.distributed.models.http import (
    Request,
    Response
)
from pydantic import BaseModel
from typing import (
    Dict, 
    Union,
    Tuple,
    Callable
)


class ZStandardCompressor(Middleware):

    def __init__(
        self, 
        serializers: Dict[
            str,
            Callable[
                [
                    Union[
                        Response,
                        BaseModel,
                        str,
                        None
                    ]
                ],
                Union[
                    str,
                    None
                ]
            ]
        ]={}
    ) -> None:
        super().__init__(
            self.__class__.__name__,
            middleware_type=MiddlewareType.UNIDIRECTIONAL_AFTER
        )

        self.serializers = serializers
        self._compressor = zstandard.ZstdCompressor()

    async def __run__(
        self, 
        request: Request,
        response: Union[
            Response,
            BaseModel,
            str,
            None
        ],
        status: int
    ) -> Tuple[
        Tuple[Response, int], 
        bool
    ]:
        try:

            if response is None:
                return (
                    Response(
                        request.path,
                        request.method,
                        data=response
                    ),
                    status
                ), True
            
            elif isinstance(response, str):

                compressed_data: bytes = self._compressor.compress(
                    response.encode()
                )

                return (
                    Response(
                        request.path,
                        request.method,
                        headers={
                            'x-compression-encoding': 'zstd',
                            'content-type': 'text/plain'
                        },
                        data=b64encode(compressed_data).decode()
                    ),
                    status
                ), True
            
            else:

                serialized = self.serializers[request.path](response)
                compressed_data: bytes = self._compressor.compress(
                    serialized
                )

                response.headers.update({
                    'x-compression-encoding': 'gzip',
                    'content-type': 'text/plain'
                })

                return (
                    Response(
                        request.path,
                        request.method,
                        headers=response.headers,
                        data=b64encode(compressed_data).decode()
                    ),
                    status
                ), True
            
        except KeyError:
            return (
                Response(
                    request.path,
                    request.method,
                    data=f'No serializer for {request.path} found.'
                ),
                500
            ), False

        except Exception as e:
            
            return (
                Response(
                    request.path,
                    request.method,
                    data=str(e)
                ),
                500
            ), False


import asyncio
import math
import random
import time
from collections import defaultdict, deque
from hedra.distributed.env import (
    Env, 
    MonitorEnv,
    load_env
)
from hedra.distributed.env.time_parser import TimeParser
from hedra.distributed.hooks.client_hook import client
from hedra.distributed.hooks.server_hook import server
from hedra.distributed.models.raft import HealthCheck, HealthStatus
from hedra.distributed.service.controller import Controller
from hedra.distributed.snowflake import Snowflake
from hedra.distributed.types import Call
from hedra.logging import (
    HedraLogger,
    logging_manager
)
from hedra.tools.helpers import cancel
from typing import Optional, Dict, Tuple, List, Deque, Union


class Monitor(Controller):

    def __init__(
        self,
        host: str,
        port: int,
        env: Optional[Env]=None,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        logs_directory: Optional[str]=None,
        workers: int=0,
    ) -> None:
        
        if workers <= 1:
            engine = 'async'

        else:
            engine = 'process'

        if env is None:
            env: Env = load_env(Env)

        if logs_directory is None:
            logs_directory = env.MERCURY_SYNC_LOGS_DIRECTORY
            
        monitor_env: MonitorEnv = load_env(MonitorEnv)

        super().__init__(
            host,
            port,
            cert_path=cert_path,
            key_path=key_path,
            workers=workers,
            env=env,
            engine=engine
        )

        self.status: HealthStatus = 'initializing'
        

        self.error_context: Optional[str] = None

        self.registration_timeout = TimeParser(
            monitor_env.MERCURY_SYNC_REGISTRATION_TIMEOUT
        ).time

        self.boot_wait = TimeParser(
            monitor_env.MERCURY_SYNC_BOOT_WAIT
        ).time
        
        self._healthcheck_task: Union[asyncio.Task, None] = None
        self._registered: Dict[int, Tuple[str, int]] = {}
        self._running = False

        self._cleanup_interval = TimeParser(
            env.MERCURY_SYNC_CLEANUP_INTERVAL
        ).time

        self._poll_interval = TimeParser(
            monitor_env.MERCURY_SYNC_HEALTH_POLL_INTERVAL
        ).time

        self._poll_timeout = TimeParser(
            monitor_env.MERCURY_SYNC_HEALTH_CHECK_TIMEOUT
        ).time

        self._local_health_multipliers: Dict[
            Tuple[str, int],
            float
        ] = defaultdict(lambda: 0)

        self._reboot_timeout = TimeParser(
            monitor_env.MERCURY_SYNC_IDLE_REBOOT_TIMEOUT
        ).time

        self._max_time_idle = TimeParser(
            monitor_env.MERCURY_SYNC_MAX_TIME_IDLE
        ).time

        self._poll_retries = monitor_env.MERCURY_SYNC_MAX_POLL_MULTIPLIER

        self._sync_interval = TimeParser(
            monitor_env.MERCURY_SYNC_UDP_SYNC_INTERVAL
        ).time

        self._suspect_max_age= TimeParser(
            monitor_env.MERCURY_SYNC_SUSPECT_MAX_AGE
        ).time

        self._check_nodes_count = monitor_env.MERCURY_SYNC_INDIRECT_CHECK_NODES

        self.min_suspect_multiplier = monitor_env.MERCURY_SYNC_MIN_SUSPECT_TIMEOUT_MULTIPLIER
        self.max_suspect_multiplier = monitor_env.MERCURY_SYNC_MAX_SUSPECT_TIMEOUT_MULTIPLIER 
        self._min_suspect_node_count = monitor_env.MERCURY_SYNC_MIN_SUSPECT_NODES_THRESHOLD
        self._max_poll_multiplier = monitor_env.MERCURY_SYNC_MAX_POLL_MULTIPLIER
        self._initial_expected_nodes = monitor_env.MERCURY_SYNC_EXPECTED_NODES

        self._confirmed_suspicions: Dict[Tuple[str, int], int] = defaultdict(lambda: 0)
        self._registered_counts: Dict[Tuple[str, int], int] = defaultdict(lambda: 0)
        self._waiter: Union[asyncio.Future, None] = None

        self._tasks_queue: Deque[asyncio.Task] = deque()
        self._degraded_nodes: Deque[Tuple[str, int]] = deque()
        self._suspect_nodes: Deque[Tuple[str, int]] = deque()
        self._suspect_history: List[Tuple[str, int, int]] = []

        self._degraded_tasks: Dict[Tuple[str, int], asyncio.Task] = {}
        self._suspect_tasks: Dict[Tuple[str, int], asyncio.Task] = {}
        self._latest_update: Dict[Tuple[str,int], int] = {}

        self._local_health_monitor: Union[asyncio.Task, None] = None
        self._udp_sync_task: Union[asyncio.Task, None] = None
        self._tcp_sync_task: Union[asyncio.Task, None] = None

        self._cleanup_task: Union[asyncio.Task, None] = None
        self._investigating_nodes: Dict[Tuple[str, int], Dict[Tuple[str, int]]] = defaultdict(dict)
        self._node_statuses: Dict[Tuple[str, int], HealthStatus] = {}
        self._instance_ids: Dict[Tuple[str, int], int] = {}

        self._models = [
            HealthCheck
        ]

        self.bootstrap_host: Union[str, None] = None
        self.bootstrap_port: Union[int, None] = None
        
        logging_manager.logfiles_directory = logs_directory
        logging_manager.update_log_level(
            env.MERCURY_SYNC_LOG_LEVEL
        )

        self._logger = HedraLogger()
        self._logger.initialize()

        self._healthy_statuses = [
            'initializing',
            'waiting',
            'healthy'
        ]

        self._unhealthy_statuses = [
            'suspect',
            'failed'
        ]

        self.failed_nodes: List[Tuple[str, int, float]] = []
        self.removed_nodes: List[Tuple[str, int, float]] = []

        self._failed_max_age = TimeParser(
            monitor_env.MERCURY_SYNC_FAILED_NODES_MAX_AGE
        ).time

        self._removed_max_age = TimeParser(
            monitor_env.MERCURY_SYNC_REMOVED_NODES_MAX_AGE
        ).time

    @server()
    async def register_node(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        try:
            source_host = healthcheck.source_host
            source_port = healthcheck.source_port

            not_self = self._check_is_not_self(
                source_host,
                source_port
            )

            not_registered = self._check_is_not_registered(
                source_host,
                source_port
            )

            if not_self and not_registered: 
                self._node_statuses[(source_host, source_port)] = 'healthy'

            snowflake = Snowflake.parse(shard_id)
            self._instance_ids[(source_host, source_port)] = snowflake.instance 

            if healthcheck.registered_nodes:

                for host, port, instance_id in healthcheck.registered_nodes:

                    not_self = self._check_is_not_self(
                        host,
                        port
                    )

                    not_registered = self._check_is_not_registered(
                        host,
                        port
                    )

                    if not_self and not_registered:
                        self._node_statuses[(host, port)] = 'healthy'

                        self._tasks_queue.append(
                            asyncio.create_task(
                                self._cancel_suspicion_probe(
                                    host,
                                    port
                                )
                            )
                        )
                    
                    self._instance_ids[(host, port)] = instance_id

            node_address = (source_host, source_port)

            self._tasks_queue.append(
                asyncio.create_task(
                    self._cancel_suspicion_probe(
                        source_host,
                        source_port
                    )
                )
            )

            if node_address in self.failed_nodes:
                self.failed_nodes.remove(node_address)

            self._registered_counts[(source_host, source_port)] = max(
                healthcheck.registered_count,
                self._registered_counts[(source_host, source_port)]
            )

            return HealthCheck(
                host=source_host,
                port=source_port,
                source_host=self.host,
                source_port=self.port,
                registered_nodes=[
                    (
                        host,
                        port,
                        self._instance_ids.get((
                            host, port
                        ))
                    ) for host, port in self._instance_ids
                ],
                status=self.status,
                registered_count=len(self._instance_ids)
            )

        except Exception:
            pass

    @server()
    async def deregister_node(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port

        node = self._node_statuses.get((
            source_host,
            source_port
        ))

        await self._logger.distributed.aio.info(f'Node - {source_host}:{source_port} - submitted request to leave to source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {source_host}:{source_port} - submitted request to leave to source - {self.host}:{self.port}')

        if self._suspect_tasks.get((
            source_host,
            source_port
        )):

            self._tasks_queue.append(
                asyncio.create_task(
                    self._cancel_suspicion_probe(
                        source_host,
                        source_port
                    )
                )
            )
            
            await self._logger.distributed.aio.debug(f'Source - {self.host}:{self.port} - has cancelled suspicion of node - {source_host}:{source_port} - due to leave request')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - has cancelled suspicion of node - {source_host}:{source_port} - due to leave request')

        if node is not None:
            
            node_status = "inactive"
            self._node_statuses[(source_host, source_port)] = node_status
            
            await self._logger.distributed.aio.debug(f'Source - {self.host}:{self.port} - has accepted request to remove node - {source_host}:{source_port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - has accepted request to remove node - {source_host}:{source_port}')


        return HealthCheck(
            host=healthcheck.source_host,
            port=healthcheck.source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )

    @server()
    async def update_node_status(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        update_node_host = healthcheck.source_host
        update_node_port = healthcheck.source_port
        update_status = healthcheck.status

        await self._logger.distributed.aio.debug(f'Node - {update_node_host}:{update_node_port} - updating status to - {update_status} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {update_node_host}:{update_node_port} - updating status to - {update_status} - for source - {self.host}:{self.port}')

        if healthcheck.target_host and healthcheck.target_port:
            update_node_host = healthcheck.target_host
            update_node_port = healthcheck.target_port

        if healthcheck.target_status:
            update_status = healthcheck.target_status

        target_last_updated: Union[int, None] = healthcheck.target_last_updated
        local_last_updated: Union[int, None] = self._latest_update.get((
            update_node_host,
            update_node_port
        ), 0)

        snowflake = Snowflake.parse(shard_id)

        source_host = healthcheck.source_host
        source_port = healthcheck.source_port
        self._instance_ids[(source_host, source_port)] = snowflake.instance

        if target_last_updated > local_last_updated:
            self._node_statuses[(update_node_host, update_node_port)] = update_status

        self._local_health_multipliers[(update_node_host, update_node_port)] = self._reduce_health_multiplier(
            update_node_host,
            update_node_port
        )
        
        await self._logger.distributed.aio.debug(f'Node - {update_node_host}:{update_node_port} - updated status to - {update_status} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {update_node_host}:{update_node_port} - updated status to - {update_status} - for source - {self.host}:{self.port}')

        return HealthCheck(
            host=healthcheck.source_host,
            port=healthcheck.source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )

    @server()
    async def update_as_suspect(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port

        await self._logger.distributed.aio.debug(f'Node - {source_host}:{source_port} - requested a check for suspect source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {source_host}:{source_port} - requested a check for suspect source - {self.host}:{self.port}')

        if self.status == 'healthy':

            await self._logger.distributed.aio.debug(f'Source - {self.host}:{self.port} - received notification it is suspect despite being healthy from node - {source_host}:{source_port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Source - {self.host}:{self.port} - received notification it is suspect despite being healthy from node - {source_host}:{source_port}')

            self._local_health_multipliers[(source_host, source_port)] = self._increase_health_multiplier(
                source_host,
                source_port
            )

            self._tasks_queue.append(
                asyncio.create_task(
                    self._run_healthcheck(
                        source_host,
                        source_port
                    )
                )
            )

        return HealthCheck(
            host=source_host,
            port=source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )

    @server()
    async def send_indirect_check(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port

        target_host = healthcheck.target_host
        target_port = healthcheck.target_port

        await self._logger.distributed.aio.debug(f'Node - {source_host}:{source_port} - requested an indirect check for node - {target_host}:{target_port} - from source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {source_host}:{source_port} - requested an indirect check for node - {target_host}:{target_port} - from source - {self.host}:{self.port}')

        try:

            investigation_update = self._acknowledge_indirect_probe(
                source_host,
                source_port,
                target_host,
                target_port
            )

            indirect_probe = self._run_healthcheck(
                target_host,
                target_port
            )

            for task in asyncio.as_completed([
                investigation_update,
                indirect_probe
            ]):
                await task

                self._local_health_multipliers[(target_host, target_port)] = self._reduce_health_multiplier(
                    target_host,
                    target_port
                )
                
            await self._logger.distributed.aio.debug(f'Suspect node - {target_host}:{target_port} - responded to an indirect check from source - {self.host}:{self.port} - for node - {source_host}:{source_port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Suspect node - {target_host}:{target_port} - responded to an indirect check from source - {self.host}:{self.port} - for node - {source_host}:{source_port}')

        except Exception:

            if self._node_statuses[(target_host, target_port)] != 'failed':

                await self._logger.distributed.aio.debug(f'Suspect node - {target_host}:{target_port} - failed to respond to an indirect check from source - {self.host}:{self.port} - for node - {source_host}:{source_port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Suspect node - {target_host}:{target_port} - failed to respond to an indirect check from source - {self.host}:{self.port} - for node - {source_host}:{source_port}')

                self._local_health_multipliers[(target_host, target_port)] = self._increase_health_multiplier(
                    target_host,
                    target_port
                )

                # Our suspicion is correct!
                return HealthCheck(
                    host=healthcheck.source_host,
                    port=healthcheck.source_port,
                    source_host=target_host,
                    source_port=target_port,
                    target_status='suspect', 
                    status=self.status
                )
            
        return HealthCheck(
            host=healthcheck.source_host,
            port=healthcheck.source_port,
            target_status=self._node_statuses.get((target_host, target_port)),   
            source_host=target_host,
            source_port=target_port,
            status=self.status,
            error=self.error_context
        )

    @server()
    async def update_acknowledged(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port
        target_host = healthcheck.target_host
        target_port = healthcheck.target_port


        await self._logger.distributed.aio.debug(f'Node - {source_host}:{source_port} - acknowledged the indirect check request for node - {target_host}:{target_port} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {source_host}:{source_port} - acknowledged the indirect check request for node - {target_host}:{target_port} - for source - {self.host}:{self.port}')

        if self._investigating_nodes.get((target_host, target_port)) is None:
            self._investigating_nodes[(target_host, target_port)] = {}

        self._investigating_nodes[(target_host, target_port)].update({
            (source_host, source_port): healthcheck.status
        })

        return HealthCheck(
            host=source_host,
            port=source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )
        
    @server()
    async def update_node_health(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
            
        try:
            update_node_host = healthcheck.source_host
            update_node_port = healthcheck.source_port

            local_node_status = self._node_statuses.get((update_node_host, update_node_port))

            if self._suspect_tasks.get((
                update_node_host,
                update_node_port
            )):

                await self._logger.distributed.aio.debug(f'Node - {update_node_host}:{update_node_port} - submitted healthy status to source - {self.host}:{self.port} - and is no longer suspect')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {update_node_host}:{update_node_port} - submitted healthy status to source - {self.host}:{self.port} - and is no longer suspect')

                self._tasks_queue.append(
                    asyncio.create_task(
                        self._cancel_suspicion_probe(
                            update_node_host,
                            update_node_port
                        )
                    )
                )

            snowflake = Snowflake.parse(shard_id)
            
            self._node_statuses[(update_node_host, update_node_port)] = healthcheck.status
            self._latest_update[(update_node_host, update_node_port)] = snowflake.timestamp

            return HealthCheck(
                host=healthcheck.source_host,
                port=healthcheck.source_port,
                source_host=self.host,
                source_port=self.port,
                source_status=local_node_status,
                error=self.error_context,
                status=self.status
            )
        
        except Exception:
            return HealthCheck(
                host=healthcheck.source_host,
                port=healthcheck.source_port,
                source_host=self.host,
                source_port=self.port,
                source_status=local_node_status,
                error=self.error_context,
                status=self.status
            )
    
    @client('register_node')
    async def submit_registration(
        self,
        host: str,
        port: int
    ) -> Call[HealthCheck]:    
        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            registered_nodes=[
                (
                    host,
                    port,
                    self._instance_ids.get((
                        host,
                        port
                    ))
                ) for host, port in self._instance_ids
            ],
            registered_count=len(self._instance_ids),
            error=self.error_context,
            status=self.status
        )

    @client('update_node_health')
    async def push_health_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthCheck, None] = None
        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))

        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_status=target_status,
            error=error_context,
            status=health_status
        )
    
    @client('update_node_health', as_tcp=True)
    async def push_tcp_health_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthCheck, None] = None
        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))

        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_status=target_status,
            error=error_context,
            status=health_status
        )


    
    async def _cancel_suspicion_probe(
        self,
        suspect_host: str,
        suspect_port: int
    ):
        
        suspect_node = (suspect_host, suspect_port)

        suspect_tasks = dict(self._suspect_tasks)
        suspect_task = suspect_tasks.get(suspect_node)

        if suspect_task is not None:
            await cancel(suspect_task)
            del suspect_tasks[suspect_node]
            
        self._suspect_tasks = suspect_tasks

    async def _run_tcp_healthcheck(
        self, 
        host: str, 
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None
    ) -> Union[Tuple[int, HealthCheck], None]:
        
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None

        await self._logger.distributed.aio.debug(f'Running TCP healthcheck for node - {host}:{port} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Running TCP healthcheck for node - {host}:{port} - for source - {self.host}:{self.port}')

        for idx in range(self._poll_retries):

            try:

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_tcp_health_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=self._calculate_current_timeout(
                        host,
                        port
                    )
                )

                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                self._node_statuses[(source_host, source_port)] = healthcheck.status

                self._local_health_multipliers[(host, port)] = self._reduce_health_multiplier(
                    host,
                    port
                )

                return shard_id, healthcheck

            except Exception:
                self._local_health_multipliers[(host, port)] = self._increase_health_multiplier(
                    host,
                    port
                )
                
        check_host = host
        check_port = port

        if target_host and target_port:
            check_host = target_host
            check_port = target_port

        node_status = self._node_statuses.get((
            check_host, 
            check_port
        )) 

        not_self = self._check_is_not_self(
            check_host,
            check_port
        )

        if not_self and healthcheck is None and node_status == 'healthy':

            await self._logger.distributed.aio.debug(f'Node - {check_host}:{check_port} - failed to respond over - {self._poll_retries} - retries and is now suspect for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {check_host}:{check_port} - failed to respond over - {self._poll_retries} - retries and is now suspect for source - {self.host}:{self.port}')

            self._node_statuses[(check_host, check_port)] = 'suspect'

            self._suspect_nodes.append((
                check_host,
                check_port
            ))

            self._suspect_tasks[(host, port)] = asyncio.create_task(
                self._start_suspect_monitor()
            )

        else:
            await self._logger.distributed.aio.debug(f'Node - {check_host}:{check_port} - responded on try - {idx}/{self._poll_retries} - for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {check_host}:{check_port} - responded on try - {idx}/{self._poll_retries} - for source - {self.host}:{self.port}')

        return shard_id, healthcheck
    
    @client('update_acknowledged')
    async def push_acknowledge_check(
        self,
        host: str,
        port: int,
        target_host: str,
        target_port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            status=health_status,
            error=error_context
        )
    
    @client('send_indirect_check')
    async def request_indirect_check(
        self,
        host: str,
        port: int,
        target_host: str,
        target_port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None

    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            target_host=target_host,
            target_port=target_port,
            target_status=self._node_statuses[(target_host, target_port)],
            source_host=self.host,
            source_port=self.port,
            error=error_context,
            status=health_status
        )
    
    @client('update_node_status')
    async def push_status_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthStatus, None] = None
        target_last_updated: Union[int,  None] = self._latest_update.get(
            (host, port), 0
        )

        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))
            target_last_updated = self._latest_update.get(
                (target_host, target_port), 0
            )

        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_last_updated=target_last_updated,
            target_status=target_status,
            status=health_status,
            error=error_context
        )
    
    @client('update_node_status', as_tcp=True)
    async def push_tcp_status_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthStatus, None] = None
        target_last_updated: Union[int,  None] = self._latest_update.get(
            (host, port), 0
        )

        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))
            target_last_updated = self._latest_update.get(
                (target_host, target_port), 0
            )


        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_status=target_status,
            target_last_updated=target_last_updated,
            status=health_status,
            error=error_context
        )
    
    @client('update_as_suspect')
    async def push_suspect_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=health_status,
            error=error_context
        )
    
    @client('deregister_node')
    async def request_deregistration(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=health_status,
            error=error_context
        )
    
    async def start(self):

        await self._logger.filesystem.aio.create_logfile(f'hedra.distributed.{self._instance_id}.log')
        self._logger.filesystem.create_filelogger(f'hedra.distributed.{self._instance_id}.log')

        await self.start_server()

        boot_wait = random.uniform(0.1, self.boot_wait * self._initial_expected_nodes)
        await asyncio.sleep(boot_wait)

    async def register(
        self,
        host: str,
        port: int
    ):
    
        await self._logger.distributed.aio.info(f'Initializing node - {self.host}:{self.port} - with id - {self._instance_id}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Initializing node - {self.host}:{self.port} - with id - {self._instance_id}')
  
        self.bootstrap_host = host
        self.bootstrap_port = port
        self.status = 'healthy'

        await self._logger.distributed.aio.info(f'Connecting to node node - {self.bootstrap_host}:{self.bootstrap_port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Connecting to node node - {self.bootstrap_host}:{self.bootstrap_port}')
        
        await self._register_initial_node()
        
        self._running = True
        
        
        self._healthcheck_task = asyncio.create_task(
            self.start_health_monitor()
        )
        
        self._cleanup_task = asyncio.create_task(
            self.cleanup_pending_checks()
        )

        self._udp_sync_task = asyncio.create_task(
            self._run_udp_state_sync()
        )

        self._tcp_sync_task = asyncio.create_task(
            self._run_tcp_state_sync()
        )

        await self._logger.distributed.aio.info(f'Initialized node - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Initialized node - {self.host}:{self.port}')


        self.status = 'healthy'

    async def _register_initial_node(self):
        await self._logger.distributed.aio.info(f'Connecting to initial node - {self.bootstrap_host}:{self.bootstrap_port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Connecting to initial node - {self.bootstrap_host}:{self.bootstrap_port}')

        poll_timeout = self._poll_timeout * self._initial_expected_nodes

        try:

            self._node_statuses[(self.bootstrap_host, self.bootstrap_port)] = 'healthy'
            
            await asyncio.wait_for(
                self.start_client(
                    {
                        (self.bootstrap_host, self.bootstrap_port): self._models
                    },
                    cert_path=self.cert_path,
                    key_path=self.key_path
                ),
                timeout=poll_timeout
            )

            while len(self._node_statuses) < 1:

                try:

                    shard_id, response = await asyncio.wait_for(
                        self.submit_registration(
                            self.bootstrap_host,
                            self.bootstrap_port
                        ),
                        timeout=poll_timeout
                    )

                    source_host = response.source_host
                    source_port = response.source_port

                    self._instance_ids[(source_host, source_port)] = Snowflake.parse(shard_id).instance

                except Exception:
                    pass

                await asyncio.sleep(self._poll_interval)

        except Exception:
            pass

    def _calculate_min_suspect_timeout(
        self,
        suspect_node_address: Tuple[str, int]
    ):
        nodes_count = len(self._node_statuses) + 1

        suspect_host, suspect_port = suspect_node_address

        poll_timeout = self._calculate_current_timeout(
            suspect_host,
            suspect_port
        )

        return round(
            self.min_suspect_multiplier * math.log10(nodes_count) * poll_timeout,
            2
        )
    
    def _reduce_health_multiplier(
        self,
        host: str,
        port: int
    ) -> int:
        
        modifier = len([
            address for address, status in self._node_statuses.items() if status == 'healthy' 
        ])

        return max(
            self._local_health_multipliers[(host, port)] - (1 * modifier),
            0
        )
    
    def _increase_health_multiplier(
        self,
        host: str,
        port: int
    ) -> int:
        
        return min(
            self._local_health_multipliers[(host, port)] + 1,
            self.max_suspect_multiplier
        )
    
    def _calculate_current_timeout(
        self,
        host: str,
        port: int
    ):
        modifier = max(
            len([
                address for address, status in self._node_statuses.items() if status == 'healthy' 
            ]),
            self._initial_expected_nodes
        )

        return self._poll_timeout + (self._local_health_multipliers[(host, port)] + 1) * modifier
    
    def _calculate_current_poll_interval(
        self,
        host: str,
        port: int
    ) -> float:
        return self._poll_interval * (self._local_health_multipliers[(host, port)] + 1)

    def _calculate_max_suspect_timeout(self, min_suspect_timeout: float):
        
        return round(
            self.max_suspect_multiplier * min_suspect_timeout,
            2
        )

    def _calculate_suspicion_timeout(
        self,
        suspect_node_address: Tuple[str, int]
    ):

        min_suspect_timeout = self._calculate_min_suspect_timeout(suspect_node_address)
        
        max_suspect_timeout = self._calculate_max_suspect_timeout(min_suspect_timeout)

        confirmed_suspect_count = max(
            0,
            self._confirmed_suspicions[suspect_node_address]
        )

        timeout_modifier = math.log(
            confirmed_suspect_count + 1
        )/math.log(self._min_suspect_node_count + 1)

        timeout_difference = max_suspect_timeout - min_suspect_timeout

        return max(
            min_suspect_timeout,
            max_suspect_timeout - (timeout_difference * timeout_modifier)
        )
    
    def _check_is_not_self(
        self,
        host: str,
        port: int
    ):
        return host != self.host and port != self.port
    
    def _check_is_not_registered(
        self,
        host: str,
        port: int
    ):
        return self._node_statuses.get((
            host,
            port
        )) is None
    
    async def _acknowledge_indirect_probe(
        self,
        host: str,
        port: int,
        target_host: str,
        target_port: int
    ):
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None
        
        await self._logger.distributed.aio.debug(f'Running UDP healthcheck for node - {host}:{port} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Running UDP healthcheck for node - {host}:{port} - for source - {self.host}:{self.port}')

        for idx in range(self._poll_retries):

            try:

                await self._logger.distributed.aio.debug(f'Sending indirect check request to - {target_host}:{target_port} -for node - {host}:{port} - from source - {self.host}:{self.port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Sending indirect check request to - {target_host}:{target_port} -for node - {host}:{port} - from source - {self.host}:{self.port}')

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_acknowledge_check(
                        host,
                        port,
                        target_host,
                        target_port,
                        self.status,
                        error_context=self.error_context
                    ),
                    timeout=self._calculate_current_timeout(
                        host,
                        port
                    )
                )

                shard_id, healthcheck = response

                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                not_self = self._check_is_not_self(
                    source_host,
                    source_port
                )

                if not_self:
                    self._node_statuses[(source_host, source_port)] = healthcheck.status

                await self._logger.distributed.aio.debug(f'Completed indirect check request to - {target_host}:{target_port} -for node - {host}:{port} - from source - {self.host}:{self.port} - on try - {idx}/{self._poll_retries}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Completed indirect check request to - {target_host}:{target_port} -for node - {host}:{port} - from source - {self.host}:{self.port} - on try - {idx}/{self._poll_retries}')

                return shard_id, healthcheck

            except Exception:
                pass              

    async def _run_healthcheck(
        self, 
        host: str, 
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None
    ) -> Union[Tuple[int, HealthCheck], None]:
        
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None
        
        await self._logger.distributed.aio.debug(f'Running UDP healthcheck for node - {host}:{port} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Running UDP healthcheck for node - {host}:{port} - for source - {self.host}:{self.port}')

        for idx in range(self._poll_retries):

            timeout = self._calculate_current_timeout(
                host,
                port
            )

            try:

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_health_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=timeout
                )

                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                not_self = self._check_is_not_self(
                    source_host,
                    source_port
                )

                if not_self:
                    self._node_statuses[(source_host, source_port)] = healthcheck.status

                self._local_health_multipliers[(host, port)] = self._reduce_health_multiplier(
                    host,
                    port
                )

                await self._logger.distributed.aio.debug(f'Node - {host}:{port} - responded on try - {idx}/{self._poll_retries} - for source - {self.host}:{self.port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {host}:{port} - responded on try - {idx}/{self._poll_retries} - for source - {self.host}:{self.port}')

                return shard_id, healthcheck

            except Exception:

                await self._logger.distributed.aio.debug(f'Node - {host}:{port} - failed for source node - {self.host}:{self.port} - on attempt - {idx}/{self._poll_retries}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Node - {host}:{port} - failed for source node - {self.host}:{self.port} - on attempt - {idx}/{self._poll_retries}')

                self._local_health_multipliers[(host, port)] = self._increase_health_multiplier(
                    host,
                    port
                )

        check_host = host
        check_port = port

        if target_host and target_port:
            check_host = target_host
            check_port = target_port

        node_status = self._node_statuses.get((
            check_host, 
            check_port
        )) 

        not_self = self._check_is_not_self(
            check_host,
            check_port
        )

        if not_self and healthcheck is None and node_status== 'healthy':

            await self._logger.distributed.aio.debug(f'Node - {check_host}:{check_port} - failed to respond over - {self._poll_retries} - retries and is now suspect for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {check_host}:{check_port} - failed to respond over - {self._poll_retries} - retries and is now suspect for source - {self.host}:{self.port}')

            self._node_statuses[(check_host, check_port)] = 'suspect'

            self._suspect_nodes.append((
                check_host,
                check_port
            ))

            self._suspect_tasks[(host, port)] = asyncio.create_task(
                self._start_suspect_monitor()
            )
        
        return shard_id, healthcheck

    async def _start_suspect_monitor(self) -> Tuple[str, int]:

        if len(self._suspect_nodes) < 1:
            return
        
        address = self._suspect_nodes.pop() 
        suspect_host, suspect_port = address


        not_self = self._check_is_not_self(
            suspect_host,
            suspect_port
        )

        if not_self and address not in self._suspect_history:
            self._suspect_history.append((
                suspect_host,
                suspect_port,
                time.monotonic()
            ))

        else:
            return


        status = self._node_statuses[(suspect_host, suspect_port)] 

        if status == 'suspect':

            await self._logger.distributed.aio.debug(f'Node - {suspect_host}:{suspect_port} - marked suspect for source {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {suspect_host}:{suspect_port} - marked suspect for source {self.host}:{self.port}')
            
        suspicion_timeout = self._calculate_suspicion_timeout(address)

        elapsed = 0
        start = time.monotonic()

        while elapsed < suspicion_timeout and status == 'suspect':

            self._tasks_queue.append(
                asyncio.create_task(
                    self._push_suspect_update(
                        host=suspect_host,
                        port=suspect_port,
                        health_status=self.status,
                        error_context=self.error_context
                    )
                )
            )
                
            confirmation_members = self._get_confirmation_members((
                suspect_host,
                suspect_port
            ))

            suspect_count = await self._request_indirect_probe(
                suspect_host,
                suspect_port,
                confirmation_members
            )

            self._confirmed_suspicions[(suspect_host, suspect_port)] += max(
                0,
                suspect_count - 1
            )

            indirect_ack_count = len(self._investigating_nodes[(suspect_host, suspect_port)])

            missing_ack_count = len(confirmation_members) - indirect_ack_count

            await self._logger.distributed.aio.debug(f'Source - {self.host}:{self.port} - acknowledged - {indirect_ack_count} - indirect probes and failed to acknowledge - {missing_ack_count} - indirect probes.')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Source - {self.host}:{self.port} - acknowledged - {indirect_ack_count} - indirect probes and failed to acknowledge - {missing_ack_count} - indirect probes.')

            next_health_multiplier = self._local_health_multipliers[(suspect_host, suspect_port)] + missing_ack_count - indirect_ack_count
            if next_health_multiplier < 0:
                self._local_health_multipliers[(suspect_host, suspect_port)] = 0

            else:
                self._local_health_multipliers[(suspect_host, suspect_port)] = self._increase_health_multiplier(
                    suspect_host,
                    suspect_port
                )

            confirmation_members_count = len(confirmation_members)

            if suspect_count < confirmation_members_count:
                # We had a majority confirmation the node was healthy.
                self._investigating_nodes[(suspect_host, suspect_port)] = {}
                self._confirmed_suspicions[(suspect_host, suspect_port)] = 0

                self._node_statuses[(suspect_host, suspect_port)] = 'healthy'

                self._reduce_health_multiplier(
                    suspect_host,
                    suspect_port
                )

                await self._logger.distributed.aio.info(f'Node - {suspect_host}:{suspect_port} - successfully responded to one or more probes for source - {self.host}:{self.port}')
                await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {suspect_host}:{suspect_port} - failed to respond for source - {self.host}:{self.port}. Setting next timeout as - {suspicion_timeout}')

                break
            
            await asyncio.sleep(
                self._calculate_current_poll_interval(
                    suspect_host,
                    suspect_port
                )
            )

            
            status = self._node_statuses[(suspect_host, suspect_port)] 

            elapsed = time.monotonic() - start
            suspicion_timeout = self._calculate_suspicion_timeout(address)


            await self._logger.distributed.aio.debug(f'Node - {suspect_host}:{suspect_port} - failed to respond for source - {self.host}:{self.port}. Setting next timeout as - {suspicion_timeout}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {suspect_host}:{suspect_port} - failed to respond for source - {self.host}:{self.port}. Setting next timeout as - {suspicion_timeout}')

        if self._node_statuses[(suspect_host, suspect_port)] == 'suspect':
            self._node_statuses[(suspect_host, suspect_port)] = 'failed'

            monitors = [
                address for address, status in self._node_statuses.items() if status in self._healthy_statuses
            ]

            active_nodes_count = len(monitors)

            if active_nodes_count > 0:

                self._tasks_queue.extend([
                    asyncio.create_task(
                        self._push_state_to_node(
                            host=host,
                            port=port
                        )
                    ) for host, port in monitors
                ])

            await self._logger.distributed.aio.info(f'Node - {suspect_host}:{suspect_port} - marked failed for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Node - {suspect_host}:{suspect_port} - marked failed for source - {self.host}:{self.port}')
        
        self._investigating_nodes[(suspect_host, suspect_port)] = {}
        self._confirmed_suspicions[(suspect_host, suspect_port)] = 0

        return (
            suspect_host,
            suspect_port
        )
        
    def _get_confirmation_members(self, suspect_address: Tuple[str, int]) -> List[Tuple[str, int]]:

        confirmation_members = [
            address for address in self._node_statuses.keys() if address != suspect_address
        ]

        confirmation_members_count = len(confirmation_members)

        if self._check_nodes_count > confirmation_members_count:
            self._check_nodes_count = confirmation_members_count
            
        confirmation_members = random.sample(
            confirmation_members, 
            self._check_nodes_count
        )

        return confirmation_members

    async def _request_indirect_probe(
        self,
        host: str,
        port: int,
        confirmation_members: List[Tuple[str, int]]
    ) -> Tuple[List[Call[HealthCheck]], int]:
        
        await self._logger.distributed.aio.debug(f'Requesting indirect check for node -  {host}:{port} - for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Requesting indirect check for node -  {host}:{port} - for source - {self.host}:{self.port}')

        if len(confirmation_members) < 1:
            requested_checks = [
                asyncio.create_task(
                    self._run_tcp_healthcheck(
                        host,
                        port
                    )
                )
            ]
        
        else:
            requested_checks = [
                asyncio.create_task(
                    self.request_indirect_check(
                        node_host,
                        node_port,
                        host,
                        port,
                        self.status,
                        error_context=self.error_context
                    )
                ) for node_host, node_port in confirmation_members
            ]

            requested_checks.append(
                asyncio.create_task(
                    self._run_tcp_healthcheck(
                        host,
                        port
                    )
                )
            )

        check_tasks: Tuple[List[asyncio.Task], List[asyncio.Task]] = await asyncio.wait(
            requested_checks, 
            timeout=self._calculate_current_timeout(
                host,
                port
            )
        )

        completed, pending = check_tasks

        results: List[Call[HealthCheck]]  = await asyncio.gather(
            *completed,
            return_exceptions=True
        )

        healthchecks = [
            result for result in results if isinstance(
                result,
                tuple
            ) and isinstance(
                result[0], 
                int
            ) and isinstance(
                result[1],
                HealthCheck
            )
        ]

        errors = [
            result for result in results if result not in healthchecks
        ]

        sorted_checks: List[Call[HealthCheck]] = list(sorted(
            healthchecks,
            key=lambda check: Snowflake.parse(check[0]).timestamp
        ))

        suspect = [
            (
                shard_id,
                check
            ) for shard_id, check in sorted_checks if check.target_status == 'suspect'
        ]

        healthy = [
            (
                shard_id,
                check
            ) for shard_id, check in sorted_checks if check.target_status == 'healthy'
        ]
            
        if len(healthy) < 1:
            suspect_count = len(suspect) + len(pending) + len(errors)

        else:
            suspect_checks: List[Call[HealthCheck]] = []
            for suspect_shard_id, suspect_check in suspect:

                newer_count = 0
                for healthy_shard_id, _ in healthy:
                    if suspect_shard_id > healthy_shard_id:
                        newer_count += 1

                if newer_count >= len(healthy):
                    suspect_checks.append((
                        suspect_shard_id,
                        suspect_check
                    ))

            suspect_count = len(suspect_checks) + len(pending) + len(errors)
        
        await asyncio.gather(*[
            cancel(pending_check) for pending_check in pending
        ], return_exceptions=True)


        await self._logger.distributed.aio.debug(f'Total of {suspect_count} nodes confirmed node - {host}:{port} - is suspect for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].info(f'Total of {suspect_count} nodes confirmed node -  {host}:{port} - is suspect for source - {self.host}:{self.port}')

        return suspect_count
    
    async def _propagate_state_update(
        self,
        target_host: str,
        target_port: int
    ):
        monitoring = [
            address for address, status in self._node_statuses.items() if status in self._healthy_statuses
        ]

        for host, port in monitoring:
            await self.push_health_update(
                host,
                port,
                self.status,
                target_host=target_host,
                target_port=target_port
            )

    async def run_forever(self):
        self._waiter = asyncio.Future()
        await self._waiter

    async def start_health_monitor(self):

        while self._running:

            monitors = list(self._node_statuses.keys())

            host: Union[str, None] = None
            port: Union[int, None] = None

            monitors_count = len(monitors)

            if monitors_count > 0:
                host, port = random.choice(monitors)

            node_status = self._node_statuses.get((host, port))
            if node_status in self._healthy_statuses:
        
                self._tasks_queue.append(
                    asyncio.create_task(
                        self._run_healthcheck(
                            host,
                            port
                        )
                    )
                )

            await asyncio.sleep(
                self._calculate_current_poll_interval(
                    host,
                    port
                )
            )

    async def leave(self):

        await self._submit_leave_requests()
        await self._shutdown()

    async def _submit_leave_requests(self):
        monitors = [
            address for address, status in self._node_statuses.items() if status in self._healthy_statuses
        ]

        if len(monitors) > 0:
            await asyncio.gather(*[ 
                asyncio.create_task(
                    self.request_deregistration(
                        host,
                        port,
                        self.status,
                        error_context=self.error_context
                    )
                ) for host, port in monitors
            ])

    async def _run_udp_state_sync(self):
        while self._running:

            monitors = [
                address for address, status in self._node_statuses.items() if status in self._healthy_statuses
            ]

            active_nodes_count = len(monitors)

            if active_nodes_count > 0:

                self._tasks_queue.extend([
                    asyncio.create_task(
                        self._push_state_to_node(
                            host=host,
                            port=port
                        )
                    ) for host, port in monitors
                ])

            await asyncio.sleep(
                self._sync_interval
            )

    async def _run_tcp_state_sync(self):

        await asyncio.sleep(
            self._sync_interval/2
        )

        while self._running:

            monitors = [
                address for address, status in self._node_statuses.items() if status in self._healthy_statuses
            ]

            active_nodes_count = len(monitors)

            if active_nodes_count > 0:

                self._tasks_queue.extend([
                    asyncio.create_task(
                        self._push_state_to_node_tcp(
                            host=host,
                            port=port
                        )
                    ) for host, port in monitors
                ])

            await asyncio.sleep(
                self._sync_interval
            )

    async def _push_state_to_node(
        self,
        host: str,
        port: int
    ):
        
        updates = [
            self._push_status_update(
                host=host,
                port=port,
                target_host=node_host,
                target_port=node_port
            ) for node_host, node_port in self._node_statuses if self._node_statuses.get((
                node_host,
                node_port
            )) == 'healthy' and host != node_host and port != node_port
        ]

        if len(updates) > 0:
            await asyncio.gather(*updates)

    async def _push_state_to_node_tcp(
        self,
        host: str,
        port: int
    ):
        updates = [
            asyncio.create_task(
                self._push_tcp_status_update(
                    host=host,
                    port=port,
                    target_host=node_host,
                    target_port=node_port
                )
            ) for node_host, node_port in self._node_statuses if self._node_statuses.get((
                node_host,
                node_port
            )) == 'healthy' and host != node_host and port != node_port
        ]
    
        if len(updates) > 0:
            await asyncio.gather(*updates)

    async def _push_status_update(
        self,
        host: str,
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None
    ) -> Tuple[
            Union[int, None], 
            Union[HealthCheck, None]
        ]:
        
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None

        await self._logger.distributed.aio.debug(f'Pushing UDP health update for source - {host}:{port} - to node - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Pushing UDP health update for source - {host}:{port} - to node - {self.host}:{self.port}')

        for _ in range(self._poll_retries):

            try:

                timeout = self._calculate_current_timeout(
                    host,
                    port
                )

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_status_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=timeout
                )

                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                not_self = self._check_is_not_self(
                    source_host,
                    source_port
                )
                
                if not_self:
                    self._node_statuses[(source_host, source_port)] = healthcheck.status

                return shard_id, healthcheck

            except Exception:

                self._local_health_multipliers[(host, port)] = self._increase_health_multiplier(
                    host,
                    port
                )

        return shard_id, healthcheck

    async def _push_tcp_status_update(
        self,
        host: str,
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None
    ):
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None

        await self._logger.distributed.aio.debug(f'Pushing TCP health update for source - {host}:{port} - to node - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Pushing TCP health update for source - {host}:{port} - to node - {self.host}:{self.port}')
        
        for _ in range(self._poll_retries):
            
            try:
                
                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_tcp_status_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=self._calculate_current_timeout(
                        host,
                        port
                    )
                )

                self._local_health_multipliers[(host, port)] = self._reduce_health_multiplier(
                    host,
                    port
                )
                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                not_self = self._check_is_not_self(
                    source_host,
                    source_port
                )

                if not_self:
                    self._node_statuses[(source_host, source_port)] = healthcheck.status

                return shard_id, healthcheck

            except Exception:

                self._local_health_multipliers[(host, port)] = self._increase_health_multiplier(
                    host,
                    port
                )

        return shard_id, healthcheck


    async def _push_suspect_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ):
        
        await self._logger.distributed.aio.debug(f'Pushing TCP health update for source - {host}:{port} - to suspect node - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Pushing TCP health update for source - {host}:{port} - to suspect node - {self.host}:{self.port}')
        
        try:
            response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                self.push_suspect_update(
                    host=host,
                    port=port,
                    health_status=health_status,
                    error_context=error_context
                ),
                timeout=self._calculate_current_timeout(
                    host,
                    port
                )
            )

            _, healthcheck = response

            not_self = self._check_is_not_self(
                host,
                port
            )

            if not_self:
                self._node_statuses[(host, port)] = healthcheck.status

        except Exception:
            pass          

    async def cleanup_pending_checks(self):

        await self._logger.distributed.aio.debug(f'Running cleanup for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Running cleanup for source - {self.host}:{self.port}')

        while self._running:

            pending_checks_count = 0

            for pending_check in list(self._tasks_queue):
                if pending_check.done() or pending_check.cancelled():
                    try:
                        await pending_check

                    except Exception:
                        pass

                    self._tasks_queue.remove(pending_check)
                    pending_checks_count += 1

            for node in list(self._suspect_history):
                _, _, age = node

                failed_elapsed = time.monotonic() - age

                if failed_elapsed >= self._suspect_max_age:
                    self._suspect_history.remove(node)


            for node in list(self.failed_nodes):

                _, _, age = node
                failed_elapsed = time.monotonic() - age
                removed_elapsed = time.monotonic() - age

                if node not in self.removed_nodes:
                    self.removed_nodes.append(node)

                if failed_elapsed >= self._failed_max_age:
                    self.failed_nodes.remove(node)

                elif removed_elapsed >= self._removed_max_age:
                    self.removed_nodes.remove(node)

            await self._logger.distributed.aio.debug(f'Cleaned up - {pending_checks_count} - for source - {self.host}:{self.port}')
            await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Cleaned up - {pending_checks_count} - for source - {self.host}:{self.port}')

            await asyncio.sleep(self._cleanup_interval)
    
    async def _shutdown(self):

        await self._logger.distributed.aio.debug(f'Shutdown requested for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Shutdown requested for source - {self.host}:{self.port}')

        self._running = False

        await asyncio.gather(*[
            cancel(check) for check in self._tasks_queue
        ], return_exceptions=True)

        if self._healthcheck_task:
            await cancel(self._healthcheck_task)

        if self._local_health_monitor:
            await cancel(self._local_health_monitor)
        
        if self._cleanup_task:
            await cancel(self._cleanup_task)
        
        if self._udp_sync_task:
            await cancel(self._udp_sync_task)

        if self._tcp_sync_task:
            await cancel(self._tcp_sync_task)

        await self.close()

        await self._logger.distributed.aio.debug(f'Shutdown complete for source - {self.host}:{self.port}')
        await self._logger.filesystem.aio[f'hedra.distributed.{self._instance_id}'].debug(f'Shutdown complete for source - {self.host}:{self.port}')

    async def soft_shutdown(self):
        await asyncio.gather(*[
            cancel(check) for check in self._tasks_queue
        ], return_exceptions=True)


            
            


    





from __future__ import annotations
import asyncio
import inspect
import traceback
from collections import defaultdict
from typing import Dict, List, Any
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.graphs.transitions.common.base_edge import BaseEdge
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.submit.submit import Submit
from hedra.core.graphs.stages.analyze.analyze import Analyze
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.reporting.system.system_metrics_set_types import MonitorGroup


class AnalyzeEdge(BaseEdge[Analyze]):

    def __init__(self, source: Analyze, destination: BaseEdge[Stage]) -> None:
        super(
            AnalyzeEdge,
            self
        ).__init__(
            source,
            destination
        )

        self.requires = [
            'setup_stage_experiment_config',
            'execute_stage_streamed_analytics',
            'execute_stage_results',
            'session_stage_monitors'
        ]
        self.provides = [
            'setup_stage_experiment_config',
            'execute_stage_streamed_analytics',
            'analyze_stage_summary_metrics',
            'session_stage_monitors',
            'analyze_stage_monitors'
        ]

        self.valid_states = [
            StageStates.EXECUTED, 
            StageStates.CHECKPOINTED, 
            StageStates.TEARDOWN_COMPLETE
        ]
        
        self.assigned_candidates = []
    
    async def transition(self):
        
        try:
            self.source.state = StageStates.ANALYZING
            submit_candidates = self.generate_submit_candidates()
            
            self.source.context.update(self.edge_data)
            
            for event in self.source.dispatcher.events_by_name.values():
                event.source.stage_instance = self.source
                event.context.update(self.edge_data)
                
                if event.source.context:
                    event.source.context.update(self.edge_data)   

            if self.edge_data['analyze_stage_has_results']:

                if self.timeout and self.skip_stage is False:
                    await asyncio.wait_for(self.source.run(), timeout=self.timeout)

                elif self.skip_stage is False:
                    await self.source.run()
            
            for provided in self.provides:
                self.edge_data[provided] = self.source.context[provided]

            self.destination.state = StageStates.ANALYZED

            if self.destination.context is None:
                self.destination.context = SimpleContext()

            if self.edge_data['analyze_stage_has_results']:

                self._update(self.destination)

                all_paths = []
                for path_set in self.all_paths.get(self.source.name, []):
                    all_paths.extend(path_set)
    
                for stage in submit_candidates.values():
                    if stage.name in all_paths and stage.state == StageStates.INITIALIZED:

                        if stage.context is None:
                            stage.context = SimpleContext()

                        self._update(stage)

                        stage.state = StageStates.ANALYZED

            self.source.state = StageStates.ANALYZED

            self.visited.append(self.source.name)

        except Exception as edge_exception:
            self.exception = edge_exception

        return None, self.destination.stage_type

    def _update(self, destination: Stage):

        for edge_name in self.history:

            history = self.history[edge_name]

            self.next_history[edge_name] = {}

            self.next_history[edge_name].update({
                key: value for key, value  in history.items() if key in self.provides
            })

        if self.skip_stage is False:

            session_stage_monitors: MonitorGroup = self.edge_data['session_stage_monitors']
            session_stage_monitors.update(
                self.edge_data['analyze_stage_monitors']
            )

            self.next_history.update({
                (self.source.name, destination.name): {
                    'analyze_stage_summary_metrics': self.edge_data.get(
                        'analyze_stage_summary_metrics', 
                        {}
                    ),
                    'session_stage_monitors': session_stage_monitors
                }
            })

    def split(self, edges: List[AnalyzeEdge]) -> None:

        analyze_stage_config: Dict[str, Any] = self.source.to_copy_dict()

        analyze_stage_copy = type(self.source.name, (Analyze, ), {})()
        
        for copied_attribute_name, copied_attribute_value in analyze_stage_config.items():
            if inspect.ismethod(copied_attribute_value) is False:
                setattr(analyze_stage_copy, copied_attribute_name, copied_attribute_value)

        user_hooks: Dict[str, Dict[str, Hook]] = defaultdict(dict)
        for hooks in registrar.all.values():
            for hook in hooks:
                if hasattr(self.source, hook.shortname) and not hasattr(Analyze, hook.shortname):
                    user_hooks[self.source.name][hook.shortname] = hook._call

        analyze_stage_copy.dispatcher = self.source.dispatcher.copy()

        for event in analyze_stage_copy.dispatcher.events_by_name.values():
            event.source.stage_instance = analyze_stage_copy

        minimum_edge_idx = min([edge.transition_idx for edge in edges])

        analyze_stage_copy.context = SimpleContext()
        for event in analyze_stage_copy.dispatcher.events_by_name.values():
            event.context = analyze_stage_copy.context 
            event.source.stage_instance.context = analyze_stage_copy.context
            event.source.context = analyze_stage_copy.context
            
            if event.source.shortname in user_hooks[analyze_stage_copy.name]:
                hook_call = user_hooks[analyze_stage_copy.name].get(event.source.shortname)

                hook_call = hook_call.__get__(analyze_stage_copy, analyze_stage_copy.__class__)
                setattr(analyze_stage_copy, event.source.shortname, hook_call)

                event.source._call = hook_call

            else:            
                event.source._call = getattr(analyze_stage_copy, event.source.shortname)
                event.source._call = event.source._call.__get__(analyze_stage_copy, analyze_stage_copy.__class__)
                setattr(analyze_stage_copy, event.source.shortname, event.source._call)
          
        self.source = analyze_stage_copy

        if minimum_edge_idx < self.transition_idx:
            self.skip_stage = True

    def generate_submit_candidates(self) -> Dict[str, Stage]:
    
        submit_stages: Dict[str, Submit] = self.stages_by_type.get(StageTypes.SUBMIT)
        path_lengths: Dict[str, int] = self.path_lengths.get(self.source.name)

        all_paths = self.all_paths.get(self.source.name, [])

        submit_candidates: Dict[str, Stage] = {}

        for stage_name, stage in submit_stages.items():
            if stage_name in all_paths:
                submit_candidates[stage_name] = stage

        selected_submit_candidates: Dict[str, Stage] = {}
        following_submit_stage_distances = [
            path_length for stage_name, path_length in path_lengths.items() if stage_name in submit_stages
        ]

        for stage_name in path_lengths.keys():
            stage_distance = path_lengths.get(stage_name)

            if stage_name in submit_candidates:

                if len(following_submit_stage_distances) > 0 and stage_distance <= min(following_submit_stage_distances):
                    selected_submit_candidates[stage_name] = submit_candidates.get(stage_name)

                elif len(following_submit_stage_distances) == 0:
                    selected_submit_candidates[stage_name] = submit_candidates.get(stage_name)
        
        return selected_submit_candidates

    def setup(self) -> None:

        raw_results = {}
        session_stage_monitors: MonitorGroup = {}

        for source_stage, destination_stage in self.history:
            
            stage_results = {}

            source_history: Dict[str, Any] = self.history.get((
                source_stage, 
                self.source.name
            ), {})

            if destination_stage == self.source.name:

                stage_results = source_history.get('execute_stage_results', {})

                stage_monitors = source_history.get('session_stage_monitors')
                if stage_monitors:
                    session_stage_monitors.update(stage_monitors)

            raw_results.update(stage_results)

        execute_stages = self.stages_by_type.get(StageTypes.EXECUTE)

        results_to_calculate = {}
        target_stages = {}
        for stage_name in raw_results.keys():
            stage = execute_stages.get(stage_name)
            all_paths = self.all_paths.get(stage_name)

            in_path = self.source.name in all_paths

            if in_path:
                stage.state = StageStates.ANALYZING
                results_to_calculate[stage_name] = raw_results.get(stage_name)
                target_stages[stage_name] = stage

        self.edge_data = {
            'session_stage_monitors': session_stage_monitors,
            'analyze_stage_raw_results': results_to_calculate,
            'analyze_stage_target_stages': target_stages,
            'analyze_stage_has_results': len(results_to_calculate) > 0
        }



import asyncio
import os
import pathlib
from urllib import request

from hedra.distributed.discovery.dns.core.cache import CacheNode
from hedra.distributed.models.dns_message import (
    DNSMessage,
    QueryType
)
from hedra.distributed.discovery.dns.core.url import URL
from hedra.distributed.discovery.dns.core.exceptions import (
    DNSError
)
from hedra.distributed.discovery.dns.core.nameservers import NameServer
from hedra.distributed.discovery.dns.core.record import (
    Record,
    RecordType, 
    RecordTypesMap
)
from hedra.distributed.discovery.dns.core.record.record_data_types import (
    CNAMERecordData,
    SOARecordData,
    NSRecordData
)
from hedra.distributed.env import (
    Env, 
    RegistrarEnv, 
    load_env
)
from hedra.distributed.env.time_parser import TimeParser
from typing import Tuple, Optional, List
from .base_resolver import BaseResolver
from .memoizer import Memoizer



class RecursiveResolver(BaseResolver):

    memoizer = Memoizer()

    def __init__(
        self, 
        host: str,
        port: int,
        instance_id: str,
        env: Env,
        cache: CacheNode = None
    ):
        super().__init__(
            host,
            port,
            instance_id,
            env,
            cache=cache
        )

        self.types_map = RecordTypesMap()
        self._nameserver_urls: List[str] = []

        registrar_env: RegistrarEnv = load_env(RegistrarEnv)

        self._maximum_tries = registrar_env.MERCURY_SYNC_RESOLVER_MAXIMUM_TRIES

    def add_nameserver(
        self,
        urls: List[str]
    ):
        self._nameserver_urls.extend(urls)

        for url in urls:

            self.cache.add(
                fqdn=url,
                record_type=RecordType.NS,
                data=NSRecordData(url)
            )

        nameserver = NameServer(urls)

        return nameserver.data
        
    def load_nameserver_cache(
        self,
        url: str='ftp://rs.internic.net/domain/named.cache',
        cache_file: str=os.path.join(
            os.getcwd(), 
            'named.cache.txt'
        ),
        timeout: Optional[int]=None
    ):
        
        if not os.path.isfile(cache_file):
            try:
                res = request.urlopen(
                    url, 
                    timeout=timeout
                )

                with open(cache_file, 'wb') as f:
                    f.write(res.read())

            except Exception:
                return

        cache_data = pathlib.Path(
            cache_file
        ).read_text().splitlines()

        for line in cache_data:
            if line.startswith(';'):
                continue
            parts = line.lower().split()
            if len(parts) < 4:
                continue

            name = parts[0].rstrip('.')
            # parts[1] (expires) is ignored
            record_type = self.types_map.types_by_name.get(
                parts[2], 
                RecordType.NONE
            )

            data_str = parts[3].rstrip('.')


            data = Record.create_rdata(
                record_type, 
                data_str
            )
            
            record = Record(
                name=name,
                record_type=record_type,
                data=data,
                ttl=-1,
            )

            self.cache.add(record=record)

    async def _query(
        self, 
        fqdn: str, 
        record_type: int,
        skip_cache: bool=False
    ) -> DNSMessage:
        
        current_try_count = 0

        return await self._query_tick(
            fqdn, 
            record_type, 
            skip_cache,
            current_try_count
        )

    def _get_matching_nameserver(self, fqdn: str):
        '''Return a generator of parent domains'''

        hosts: List[URL] = self._nameserver_urls
        empty = True

        while fqdn and empty:
            if fqdn in ('in-addr.arpa', ):
                break
            _, _, fqdn = fqdn.partition('.')

            for rec in self.cache.query(fqdn, RecordType.NS):
                record_data: NSRecordData = rec.data
                host = record_data.data

                url = URL(
                    host,
                    port=self.client.port
                )

                if url.ip_type is None:
                    # host is a hostname instead of IP address

                    for res in self.cache.query(
                        host, 
                        self.nameserver_types
                    ):
                        hosts.append(
                            URL(
                                res.data.data,
                                port=self.client.port
                            )
                        )

                        empty = False

                else:
                    hosts.append(url)
                    empty = False

        return NameServer(hosts)

    @memoizer.memoize_async(
        lambda _, fqdn, record_type, skip_cache: (fqdn, record_type)
    )
    async def _query_tick(
        self, 
        fqdn: str, 
        record_type: int,
        skip_cache: bool,
        current_try_count: int
    ):
        
        msg = DNSMessage()
        msg.query_domains.append(
            Record(
                query_type=QueryType.REQUEST, 
                name=fqdn, 
                record_type=record_type
            )
        )

        has_result = False

        if skip_cache is False:
            has_result, fqdn = self.query_cache(msg, fqdn, record_type)

        last_err = None
        nameserver = self._get_matching_nameserver(fqdn)

        while not has_result and current_try_count < self._maximum_tries:

            current_try_count += 1

            for url in nameserver.iter():
                try:
                    has_result, fqdn, nsips = await self._query_remote(
                        msg, 
                        fqdn, 
                        record_type, 
                        url,
                        current_try_count
                    )

                    nameserver = NameServer(
                        self.client.port,
                        nameservers=nsips
                    )

                except Exception as err:
                    last_err = err

                else:
                    break
            else:
                raise last_err or Exception('Unknown error')


        assert has_result, 'Maximum nested query times exceeded'

        return msg

    async def _query_remote(
        self, 
        msg: DNSMessage, 
        fqdn: str, 
        record_type: RecordType,
        url: URL,
        current_try_count: int
    ):
        
        result: DNSMessage = await self.request(
            fqdn, 
            msg, 
            url
        )

        if result.query_domains[0].name != fqdn:
            raise DNSError(-1, 'Question section mismatch')
        
        assert result.query_result_code != 2, 'Remote server fail'

        self.cache_message(result)

        has_cname = False
        has_result = False
        has_ns = False

        for rec in result.query_answers:
            msg.query_answers.append(rec)

            if isinstance(rec.data, CNAMERecordData):
                fqdn = rec.data.data
                has_cname = True

            if rec.record_type != RecordType.CNAME or record_type in (
                RecordType.ANY, 
                RecordType.CNAME
            ):
                has_result = True

        for rec in result.query_namservers:
            if rec.record_type in (
                RecordType.NS, 
                RecordType.SOA
            ):
                has_result = True

            else:
                has_ns = True

        if not has_cname and not has_ns:
            # Not found, return server fail since we are not authorative
            msg = DNSMessage(
                **msg.dict(),
                query_result_code=2
            )

            has_result = True
        if has_result:
            return has_result, fqdn, []

        # Load name server IPs from res.ar
        namespace_ip_address_map = {}

        for record in result.query_additional_records:
            if record.record_type in self.nameserver_types:
                namespace_ip_address_map[(rec.name, record.record_type)] = rec.data.data

        hosts = []
        for record in result.query_namservers:

            if isinstance(record.data, SOARecordData):
                hosts.append(record.data.mname)

            elif isinstance(record.data, NSRecordData):
                hosts.append(record.data.data)

        namespace_ips = []

        for host in hosts:
            for record_type in self.nameserver_types:
                ip = namespace_ip_address_map.get((host, record_type))

                if ip is not None:
                    namespace_ips.append(ip)

        # Usually name server IPs will be included in res.ar.
        # In case they are not, query from remote.
        if len(namespace_ips) < 1 and len(hosts) > 0:

            current_try_count += 1

            for record_type in self.nameserver_types:
                for host in hosts:
                    try:
                        query_tick_result: Tuple[DNSMessage, bool] = await asyncio.shield(
                            self._query_tick(
                                host, 
                                record_type, 
                                False,
                                current_try_count
                            )
                        )

                        (
                            ns_res, 
                            _
                        ) = query_tick_result

                    except Exception:
                        pass

                    else:
                        for rec in ns_res.query_answers:
                            if rec.record_type == record_type:
                                namespace_ips.append(rec.data.data)
                                break

                if len(namespace_ips) > 0:
                    break

        return has_result, fqdn, namespace_ips




from .command_library import CommandLibrary


class CommandLibrarian:

    __slots__ = (
        'command_library'
    )

    def __init__(self, page) -> None:
        self.command_library = CommandLibrary(page)

    def get(self, command_name):
        return self.command_library.__getattribute__(command_name)

    


import asyncio
import functools
import signal
import uuid
import psutil
from concurrent.futures import ThreadPoolExecutor
from typing import (
    List, 
    Dict, 
    Union,
    Any,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .snowflake_connector_config import SnowflakeConnectorConfig


def noop_create_engine():
    pass


try:
    import sqlalchemy
    from snowflake.sqlalchemy import URL
    from sqlalchemy import create_engine
    from sqlalchemy.engine import (
        Engine, 
        Connection
    )
    
    has_connector = True

except Exception:
    sqlalchemy = object
    URL = object
    create_engine=noop_create_engine
    Engine = object
    Connection = object
    has_connector = False


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class SnowflakeConnector:
    connector_type=ConnectorType.Snowflake

    def __init__(
        self, 
        config: SnowflakeConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.username = config.username
        self.password = config.password
        self.organization_id = config.organization_id
        self.account_id = config.account_id
        self.private_key = config.private_key
        self.warehouse = config.warehouse
        self.database = config.database
        self.schema = config.database_schema
        self.stage = stage
        self.parser_config = parser_config

        self.table_name = config.table_name

        self.connect_timeout = config.connect_timeout
        
        self.metadata = sqlalchemy.MetaData()
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self._engine: Union[Engine, None] = None
        self._connection: Union[Connection, None] = None

        self._table: Union[sqlalchemy.Table, None] = None

        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        
        self.parser = Parser()

    async def connect(self):

        try:

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop
                    )
                )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')
            self._engine = await self._loop.run_in_executor(
                self._executor,
                create_engine,
                URL(
                    user=self.username,
                    password=self.password,
                    account=self.account_id,
                    warehouse=self.warehouse,
                    database=self.database,
                    schema=self.schema
                )
            
            )

            self._connection = await asyncio.wait_for(
                self._loop.run_in_executor(
                    self._executor,
                    self._engine.connect
                ),
                timeout=self.connect_timeout
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')

        except asyncio.TimeoutError:
            raise Exception('Err. - Connection to Snowflake timed out - check your account id, username, and password.')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        
        actions: List[Dict[str, Any]] = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })

    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        if self._table is None:
            
            self._table = sqlalchemy.Table(
                self.table_name,
                self.metadata,
                autoload_with=self._engine
            )
            
        results = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._connection.execute,
                self._table.select(**options)
            )
        )
        
        all_results = await self._loop.run_in_executor(
            self._executor,
            results.fetchall
        )

        return [
            {
              column: value  for column, value in result._mapping.items()
            } for result in all_results
        ]
    
    async def close(self):
        await self._loop.run_in_executor(
            self._executor,
            self._connection.close
        )

        self._executor.shutdown()


from hedra.core.engines.types.common.types import RequestTypes
from typing import (
    Dict, 
    Callable, 
    Union,
    Iterable
)
try:
    from cassandra.cqlengine.models import Model
    has_connector = True

except Exception:
    Model = None
    has_connector = False

from .action_schemas import (
    CassandraGraphQLActionSchema,
    CassandraGraphQLHTTP2ActionSchema,
    CassandraGRPCActionSchema,
    CassandraHTTPActionSchema,
    CassandraHTTP2ActionSchema,
    CassandraHTTP3ActionSchema,
    CassandraPlaywrightActionSchema,
    CassandraTaskSchema,
    CassandraUDPActionSchema,
    CassandraWebsocketActionSchema
)

from .result_schemas import (
    CassandraGraphQLResultSchema,
    CassandraGraphQLHTTP2ResultSchema,
    CassandraGRPCResultSchema,
    CassandraHTTPResultSchema,
    CassandraHTTP2ResultSchema,
    CassandraHTTP3ResultSchema,
    CassandraPlaywrightResultSchema,
    CassandraTaskResultSchema,
    CassandraUDPResultSchema,
    CassandraWebsocketResultSchema
)


class CassandraSchemaSet:

    def __init__(self) -> None:
        self._action_schemas: Dict[
            RequestTypes,
            Callable[
                [str],
                Union[
                    CassandraGraphQLActionSchema,
                    CassandraGraphQLHTTP2ActionSchema,
                    CassandraGRPCActionSchema,
                    CassandraHTTPActionSchema,
                    CassandraHTTP2ActionSchema,
                    CassandraHTTP3ActionSchema,
                    CassandraPlaywrightActionSchema,
                    CassandraTaskSchema,
                    CassandraUDPActionSchema,
                    CassandraWebsocketActionSchema
                ]
            ]
        ] = {
            RequestTypes.GRAPHQL: lambda table_name: CassandraGraphQLActionSchema(table_name),
            RequestTypes.GRAPHQL_HTTP2: lambda table_name: CassandraGraphQLHTTP2ActionSchema(table_name),
            RequestTypes.GRPC: lambda table_name: CassandraGRPCActionSchema(table_name),
            RequestTypes.HTTP: lambda table_name: CassandraHTTPActionSchema(table_name),
            RequestTypes.HTTP2: lambda table_name: CassandraHTTP2ActionSchema(table_name),
            RequestTypes.HTTP3: lambda table_name: CassandraHTTP3ActionSchema(table_name),
            RequestTypes.PLAYWRIGHT: lambda table_name: CassandraPlaywrightActionSchema(table_name),
            RequestTypes.TASK: lambda table_name: CassandraTaskSchema(table_name),
            RequestTypes.UDP: lambda table_name: CassandraUDPActionSchema(table_name),
            RequestTypes.WEBSOCKET: lambda table_name: CassandraWebsocketActionSchema(table_name)
        }

        self._results_schemas: Dict[
            RequestTypes,
            Callable[
                [str],
                Union[
                    CassandraGraphQLResultSchema,
                    CassandraGraphQLHTTP2ResultSchema,
                    CassandraGRPCResultSchema,
                    CassandraHTTPResultSchema,
                    CassandraHTTP2ResultSchema,
                    CassandraHTTP3ResultSchema,
                    CassandraPlaywrightResultSchema,
                    CassandraTaskResultSchema,
                    CassandraUDPResultSchema,
                    CassandraWebsocketResultSchema
                ]
            ]
        ] = {
            RequestTypes.GRAPHQL: lambda table_name: CassandraGraphQLResultSchema(table_name),
            RequestTypes.GRAPHQL_HTTP2: lambda table_name: CassandraGraphQLHTTP2ResultSchema(table_name),
            RequestTypes.GRPC: lambda table_name: CassandraGRPCResultSchema(table_name),
            RequestTypes.HTTP: lambda table_name: CassandraHTTPResultSchema(table_name),
            RequestTypes.HTTP2: lambda table_name: CassandraHTTP2ResultSchema(table_name),
            RequestTypes.HTTP3: lambda table_name: CassandraHTTP3ResultSchema(table_name),
            RequestTypes.PLAYWRIGHT: lambda table_name: CassandraPlaywrightResultSchema(table_name),
            RequestTypes.TASK: lambda table_name: CassandraTaskResultSchema(table_name),
            RequestTypes.UDP: lambda table_name: CassandraUDPResultSchema(table_name),
            RequestTypes.WEBSOCKET: lambda table_name: CassandraWebsocketResultSchema(table_name)
        }

    def get_action_schema(
        self, 
        request_type: RequestTypes,
        table_name: str
    ):
        return self._action_schemas.get(
            request_type,
            CassandraHTTPActionSchema
        )(table_name)
    
    def get_result_schema(
        self, 
        request_type: RequestTypes,
        table_name: str
    ):
        return self._results_schemas.get(
            request_type,
            CassandraHTTPResultSchema
        )(table_name)

    def action_schemas(
        self, 
        table_name: str
    ):
        for schema in self._action_schemas.values():
            cassandra_schema = schema(table_name)

            request_type = cassandra_schema.type.lower()
            assembled_table_name = f'{table_name}_{request_type}'

            yield assembled_table_name, cassandra_schema.actions_table
    
    def results_schemas(
        self, 
        table_name: str
    ):
        for schema in self._results_schemas.values():
            cassandra_schema = schema(table_name)

            request_type = cassandra_schema.type.lower()
            assembled_table_name = f'{table_name}_{request_type}'

            yield assembled_table_name, cassandra_schema.results_table



import asyncio
from collections import defaultdict
from typing import (
    Dict, 
    Coroutine, 
    List, 
    Tuple, 
    Optional, 
    Callable, 
    Awaitable, 
    Any
)
from hedra.core.engines.client.time_parser import TimeParser
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.simple_context import SimpleContext


class TransformHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: Optional[Tuple[str, ...]],
        timeout: Optional[float]='1m',
        pre: bool=False,
        order: int=1,
        skip: bool=False
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            order=order,
            skip=skip,
            hook_type=HookType.TRANSFORM
        )
        
        parser = TimeParser(time_amount=timeout)
        
        self.timeout = parser.time
        self.names = list(set(names))
        self.pre = pre
        self.events: Dict[str, Coroutine] = {}
        self.context: Optional[SimpleContext] = None
        self.conditions: Optional[List[Callable[..., bool]]] = []
        self._timeout_as_string = timeout
        
    async def call(self, **kwargs):

        if self.skip:
            return kwargs
        
        batchable_args: List[Dict[str, Any]] = []
        for name, arg in kwargs.items():
            if isinstance(arg, (list, tuple)):
                batchable_args.extend([
                    {**kwargs, name: item} for item in arg
                ])

        if len(batchable_args) > 0:
 
            result = await asyncio.wait_for(
                asyncio.gather(*[
                    asyncio.create_task(         
                        self._call(**{name: value for name, value in call_kwargs.items() if name in self.params})
                    ) for call_kwargs in batchable_args if (
                        await self._execute_call(**call_kwargs) is True
                    )
                ]),
                timeout=self.timeout
            )

            aggregated_transformm = defaultdict(list)

            for data_item in result:
                if isinstance(data_item, dict):
                    for name, value in data_item.items():
                        if isinstance(value, (list, tuple)):
                            aggregated_transformm[name].extend(value)

                        else:
                            aggregated_transformm[name].append(value)

                elif data_item is not None:
                    aggregated_transformm[self.shortname].append(data_item)


            return {
                **kwargs,
                **dict(aggregated_transformm)
            }

        else:     
    
            result = await self._call(**{name: value for name, value in kwargs.items() if name in self.params})

            if isinstance(result, dict):
                return {
                    **kwargs,
                    **result
                }

            return {
                **kwargs,
                self.shortname: result
            }

    def copy(self):
        transform_hook = TransformHook(
            self.name,
            self.shortname,
            self._call,
            *self.names,
            timeout=self._timeout_as_string,
            pre=self.pre,
            order=self.order,
            skip=self.skip,
        )
        
        transform_hook.stage = self.stage

        return transform_hook


class Timeouts:

    __slots__ = (
        'connect_timeout',
        'socket_read_timeout',
        'total_timeout'
    )

    def __init__(self, connect_timeout: int=10, socket_read_timeout: int=10, total_timeout: int=60) -> None:
        self.connect_timeout = connect_timeout
        self.socket_read_timeout = socket_read_timeout
        self.total_timeout = total_timeout


import asyncio
import functools
import os
import psutil
import signal
import uuid
from typing import List, Dict
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .cassandra_config import CassandraConfig



def noop_sync_table():
    pass


try:
    from cassandra.cqlengine import columns
    from cassandra.cqlengine import connection
    from cassandra.cqlengine.management import sync_table
    from cassandra.cqlengine.models import Model
    from cassandra.cluster import Cluster
    from cassandra.auth import PlainTextAuthProvider
    has_connector = True

except Exception:
    columns = object
    connection = object
    sync_table = noop_sync_table
    Model = None
    Cluster = None
    PlainTextAuthProvider = None
    has_connector = False

def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class Cassandra:

    def __init__(self, config: CassandraConfig) -> None:
        self.cluster = None
        self.session = None

        self.hosts = config.hosts
        self.port = config.port or 9042

        self.username = config.username
        self.password = config.password
        self.keyspace = config.keyspace
        
        self.events_table_name: str = config.events_table
        self.metrics_table_name: str = config.metrics_table
        self.streams_table_name: str = config.streams_table

        self.experiments_table_name: str= config.experiments_table
        self.variants_table_name: str= f'{config.experiments_table}_variants'
        self.mutations_table_name: str= f'{config.experiments_table}_mutations'

        self.shared_metrics_table_name = f'{config.metrics_table}_shared'
        self.custom_metrics_table_name = f'{config.metrics_table}_custom'
        self.errors_table_name = f'{config.metrics_table}_errors'

        self.session_system_metrics_table_name = f'{config.system_metrics_table}_session'
        self.stage_system_metrics_table_name = f'{config.system_metrics_table}_stage'
        
        self.replication_strategy = config.replication_strategy
        self.replication = config.replication       
        self.ssl = config.ssl

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._metrics_table = None
        self._errors_table = None
        self._events_table = None
        self._streams_table = None

        self._experiments_table = None
        self._variants_table = None
        self._mutations_table = None

        self._session_system_metrics_table = None
        self._stage_system_metrics_table = None

        self._shared_metrics_table = None
        self._custom_metrics_table = None

        self._executor =ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        host_port_combinations = ', '.join([
            f'{host}:{self.port}' for host in self.hosts
        ])

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Cassandra Cluster at - {host_port_combinations}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        auth = None
        if self.username and self.password:
            auth = PlainTextAuthProvider(self.username, self.password)
        

        self.cluster = Cluster(
            self.hosts,
            port=self.port,
            auth_provider=auth,
            ssl_context=self.ssl
        )

        self.session = await self._loop.run_in_executor(
            None,
            self.cluster.connect
        )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Cassandra Cluster at - {host_port_combinations}')

        if self.keyspace is None:
            self.keyspace = 'hedra'

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Keyspace - {self.keyspace}')

        keyspace_options = f"'class' : '{self.replication_strategy}', 'replication_factor' : {self.replication}"
        keyspace_query = f"CREATE KEYSPACE IF NOT EXISTS {self.keyspace} WITH REPLICATION = " + "{" + keyspace_options  + "};"

        await self._loop.run_in_executor(
            None,
            self.session.execute,
            keyspace_query
        )

        await self._loop.run_in_executor(
            None,
            self.session.set_keyspace,
            self.keyspace
        )
        if os.getenv('CQLENG_ALLOW_SCHEMA_MANAGEMENT') is None:
            os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                connection.setup,
                self.hosts, 
                self.keyspace, 
                protocol_version=3
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Keyspace - {self.keyspace}')
    
    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to - Keyspace: {self.keyspace} - Table: {self.session_system_metrics_table_name}')
        
        if self._session_system_metrics_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Session System Metrics table - {self.session_system_metrics_table_name} - under keyspace - {self.keyspace}')

            fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'group': columns.Text(),
                'median': columns.Float(),
                'mean': columns.Float(),
                'variance': columns.Float(),
                'stdev': columns.Float(),
                'minimum': columns.Float(),
                'maximum': columns.Float(),
                'created_at': columns.DateTime(default=datetime.now)
            }

            for quantile_name in SystemMetricsSet.quantiles:
                fields[quantile_name] = columns.Float()

            self._session_system_metrics_table = type(
                self.session_system_metrics_table_name.capitalize(), 
                (Model, ), 
                fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._session_system_metrics_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Session System Metrics table - {self.session_system_metrics_table_name} - under keyspace - {self.keyspace}')
        
        rows: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                rows.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                rows.append(monitor_metrics)

        for metrics_set in rows:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.name}:{metrics_set.group}')

            await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._session_system_metrics_table.create,
                        **metrics_set.record
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.session_system_metrics_table_name}')
    
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to - Keyspace: {self.keyspace} - Table: {self.stage_system_metrics_table_name}')
        
        if self._stage_system_metrics_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Stage System Metrics table - {self.stage_system_metrics_table_name} - under keyspace - {self.keyspace}')

            fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(),
                'group': columns.Text(),
                'median': columns.Float(),
                'mean': columns.Float(),
                'variance': columns.Float(),
                'stdev': columns.Float(),
                'minimum': columns.Float(),
                'maximum': columns.Float(),
                'created_at': columns.DateTime(default=datetime.now)
            }

            for quantile_name in SystemMetricsSet.quantiles:
                fields[quantile_name] = columns.Float()

            self._stage_system_metrics_table = type(
                self.stage_system_metrics_table_name.capitalize(), 
                (Model, ), 
                fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._stage_system_metrics_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Stage System Metrics table - {self.stage_system_metrics_table_name} - under keyspace - {self.keyspace}')
        
        rows: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    rows.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    rows.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    rows.append(stage_mb_per_vu_metrics)

        for metrics_set in rows:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.name}:{metrics_set.group}')

            await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._stage_system_metrics_table.create,
                        **metrics_set.record
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.stage_system_metrics_table_name}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to - Keyspace: {self.keyspace} - Table: {self.streams_table_name}')
        
        if self._streams_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Streams table - {self.streams_table_name} - under keyspace - {self.keyspace}')

            fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(min_length=1),
                'group': columns.Text(),
                'median': columns.Float(),
                'mean': columns.Float(),
                'variance': columns.Float(),
                'stdev': columns.Float(),
                'minimum': columns.Float(),
                'maximum': columns.Float(),
                'created_at': columns.DateTime(default=datetime.now)
            }

            self._streams_table = type(
                self.streams_table_name.capitalize(), 
                (Model, ), 
                fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._streams_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Streams table - {self.streams_table_name} - under keyspace - {self.keyspace}')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped:
                await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self._streams_table.create,
                            **{
                                'name': f'{stage_name}_streams',
                                'stage': stage_name,
                                'group': group_name,
                                **group
                            }
                        )
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.streams_table_name}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to - Keyspace: {self.keyspace} - Table: {self.experiments_table_name}')

        if self._experiments_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Experiments table - {self.experiments_table_name} - under keyspace - {self.keyspace}')

            self._experiments_table = type(
                self.experiments_table_name.capitalize(), 
                (Model, ), 
                {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'experiment_name': columns.Text(min_length=1, index=True),
                    'experiment_randomized': columns.Boolean(),
                    'experiment_completed': columns.Integer(),
                    'experiment_succeeded': columns.Integer(),
                    'experiment_failed': columns.Integer(),
                    'experiment_median_aps': columns.Float(),
                    'created_at': columns.DateTime(default=datetime.now)
                }
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._experiments_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Experiments table - {self.experiments_table_name} - under keyspace - {self.keyspace}')

        for experiment in experiment_metrics.experiments:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._metrics_table.create,
                    **experiment
                )
            )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to - Keyspace: {self.keyspace} - Table: {self.experiments_table_name}')
    
    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to - Keyspace: {self.keyspace} - Table: {self.variants_table_name}')

        if self._variants_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Variants table - {self.variants_table_name} - under keyspace - {self.keyspace}')

            self._variants_table = type(
                self.variants_table_name.capitalize(), 
                (Model, ), 
                {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'variant_name': columns.Text(min_length=1, index=True),
                    'variant_experiment': columns.Text(min_length=1),
                    'variant_weight': columns.Float(),
                    'variant_distribution': columns.Text(min_length=1),
                    'variant_distribution_interval': columns.Float(),
                    'variant_ratio_completed': columns.Float(),
                    'variant_ratio_succeeded': columns.Float(),
                    'variant_ratio_failed': columns.Float(),
                    'variant_ratio_aps': columns.Float(),
                    'variant_completed': columns.Integer(),
                    'variant_succeeded': columns.Integer(),
                    'variant_failed': columns.Integer(),
                    'variant_actions_per_second': columns.Float(),
                    'created_at': columns.DateTime(default=datetime.now)
                }
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._variants_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Variants table - {self.variants_table_name} - under keyspace - {self.keyspace}')

        for variant in experiment_metrics.variants:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._metrics_table.create,
                    **variant
                )
            )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to - Keyspace: {self.keyspace} - Table: {self.variants_table_name}')
    
    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to - Keyspace: {self.keyspace} - Table: {self.mutations_table_name}')

        if self._mutations_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Mutations table - {self.mutations_table_name} - under keyspace - {self.keyspace}')

            self._mutations_table = type(
                self.mutations_table_name.capitalize(), 
                (Model, ), 
                {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'mutation_name': columns.Text(min_length=1, index=True),
                    'mutation_experiment_name': columns.Text(min_length=1),
                    'mutation_variant_name': columns.Text(min_length=1),
                    'mutation_chance': columns.Float(),
                    'mutation_targets': columns.Text(min_length=1),
                    'mutation_type': columns.Text(min_length=1),
                    'created_at': columns.DateTime(default=datetime.now)
                }
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._mutations_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Mutations table - {self.mutations_table_name} - under keyspace - {self.keyspace}')

        for mutation in experiment_metrics.mutations:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._metrics_table.create,
                    **mutation
                )
            )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to - Keyspace: {self.keyspace} - Table: {self.mutations_table_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to - Keyspace: {self.keyspace} - Table: {self.events_table_name}')

        if self._events_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - under keyspace - {self.keyspace}')

            self._events_table = type(
                self.events_table_name.capitalize(), 
                (Model, ), 
                {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'name': columns.Text(min_length=1, index=True),
                    'stage': columns.Text(min_length=1),
                    'time': columns.Float(),
                    'succeeded': columns.Boolean(),
                    'created_at': columns.DateTime(default=datetime.now)
                }
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._events_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Events table - {self.events_table_name} - under keyspace - {self.keyspace}')

        for event in events:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._metrics_table.create,
                    **event.record
                )
            )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to - Keyspace: {self.keyspace} - Table: {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.shared_metrics_table_name}')
        
        if self._shared_metrics_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - under keyspace - {self.keyspace}')

            fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(min_length=1),
                'group': columns.Text(min_length=1),
                'total': columns.Integer(),
                'succeeded': columns.Integer(),
                'failed': columns.Integer(),
                'actions_per_second': columns.Float()
            }

            self._shared_metrics_table = type(
                self.shared_metrics_table_name.capitalize(), 
                (Model, ), 
                fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._shared_metrics_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Shared Metrics table - {self.shared_metrics_table_name} - under keyspace - {self.keyspace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._metrics_table.create,
                        **{
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': 'common',
                            **metrics_set.common_stats
                        }
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
       
            if self._metrics_table is None:
                
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - under keyspace - {self.keyspace}')
                
                fields = {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'name': columns.Text(min_length=1, index=True),
                    'stage': columns.Text(min_length=1),
                    'group': columns.Text(),
                    'median': columns.Float(),
                    'mean': columns.Float(),
                    'variance': columns.Float(),
                    'stdev': columns.Float(),
                    'minimum': columns.Float(),
                    'maximum': columns.Float(),
                    'created_at': columns.DateTime(default=datetime.now)
                }


                for quantile_name in metrics_set.quantiles:
                    fields[quantile_name] = columns.Float()
                
                for custom_field_name, custom_field_type in metrics_set.custom_schemas:
                    fields[custom_field_name] = custom_field_type


                self._metrics_table = type(
                    self.metrics_table_name.capitalize(), 
                    (Model, ), 
                    fields
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        sync_table,
                        self._metrics_table,
                        keyspaces=[self.keyspace]
                    )
                )

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Metrics table - {self.metrics_table_name} - under keyspace - {self.keyspace}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')                
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._metrics_table.create,
                        **{
                            **group.record,
                            'group': group_name
                        }
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics Set to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics Set to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - Custom')

        if self._custom_metrics_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics table - {self.custom_metrics_table_name} - under keyspace - {self.keyspace}')
            
            custom_table = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(min_length=1),
                'group': columns.Text(min_length=1)
            }

            for metrics_set in metrics_sets:
                for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                    cassandra_column = columns.Integer()

                    if custom_metric.metric_type == MetricType.COUNT:
                        cassandra_column = columns.Integer()
                    
                    else:
                        cassandra_column = columns.Float()

                    custom_table[custom_metric_name] = cassandra_column
                    custom_table[custom_metric_name] = cassandra_column

            self._custom_metrics_table: Model = type(
                self.custom_metrics_table_name.capitalize(), 
                (Model, ), 
                custom_table
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._custom_metrics_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Custom Metrics table - {self.custom_metrics_table_name} - under keyspace - {self.keyspace}')


        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
                

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._custom_metrics_table.create,
                    {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'Custom',
                        **{
                            custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                        }
                    }
                )
            )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics Set to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics to - Keyspace: {self.keyspace} - Table: {self.errors_table_name}')

        if self._errors_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Errors Metrics table - {self.errors_table_name} - under keyspace - {self.keyspace}')

            errors_table_fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(min_length=1),
                'error_message': columns.Text(min_length=1),
                'error_count': columns.Integer()
            }

            self._errors_table = type(
                self.errors_table_name.capitalize()
                (Model, ),
                errors_table_fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._errors_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Errors Metrics table - {self.errors_table_name} - under keyspace - {self.keyspace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._errors_table.create,
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        }
                    )
                )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to - Keyspace: {self.keyspace} - Table: {self.errors_table_name}')
 
    async def close(self):
        host_port_combinations = ', '.join([
            f'{host}:{self.port}' for host in self.hosts
        ])

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to Cassandra Cluster at - {host_port_combinations}')

        await self._loop.run_in_executor(
            self._executor,
            self.cluster.shutdown
        )

        self._executor.shutdown(cancel_futures=True)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to Cassandra Cluster at - {host_port_combinations}')


from typing import Any, Dict, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.graphql_http2 import (
    MercuryGraphQLHTTP2Client,
    GraphQLHTTP2Action,
    GraphQLHTTP2Result
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.core.engines.types.tracing.trace_session import (
    TraceSession, 
    Trace
)
from hedra.logging import HedraLogger
from .base_client import BaseClient


class GraphQLHTTP2Client(BaseClient[MercuryGraphQLHTTP2Client, GraphQLHTTP2Action, GraphQLHTTP2Result]):

    def __init__(self, config: Config) -> None:
        super().__init__()

        if config is None:
            config = Config()

        tracing_session: Union[TraceSession, None] = None
        if config.tracing:
            trace_config_dict = config.tracing.to_dict()
            tracing_session = TraceSession(**trace_config_dict)
        
        self.session = MercuryGraphQLHTTP2Client(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections,
            tracing_session=tracing_session
        )
        self.request_type = RequestTypes.GRAPHQL_HTTP2
        self.client_type = self.request_type.capitalize()

        self.actions: ActionsStore = None
        self.next_name = None
        self.intercept = False
        self.waiter = None

        self.logger = HedraLogger()
        self.logger.initialize()

    def __getitem__(self, key: str):
        return self.session.registered.get(key)

    async def query(
        self,
        url: str, 
        query: str,
        operation_name: str = None,
        variables: Dict[str, Any] = None, 
        headers: Dict[str, str] = {}, 
        user: str = None, 
        tags: List[Dict[str, str]] = [],
        trace: Trace=None
    ):
        if trace and self.session.tracing_session is None:
            self.session.tracing_session = TraceSession(
                **trace.to_dict()
            )

        request = GraphQLHTTP2Action(
            self.next_name,
            url,
            method='POST',
            headers=headers,
            data={
                "query": query,
                "operation_name": operation_name,
                "variables": variables
            },
            user=user,
            tags=tags
        )

        return await self._execute_action(request)
        



from hedra.core.graphs.stages import (
    Execute,
)

from hedra.core.hooks import (
    task
)


class ExecuteTaskStage(Execute):

    counter=0

    @task()
    async def task_http_get(self):

        response = await self.client.http.get('https://<url_here>')

        if response.status >= 200 and response.status < 300:
            self.counter += 1

        return response


from typing import Optional, Dict, Union, Tuple
from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictInt, 
    StrictFloat,
    StrictBool, 
    validator
)


class TaskHookValidator(BaseModel):
    names: Tuple[StrictStr, ...]
    weight: StrictInt
    order: StrictInt
    skip: StrictBool
    metadata: Optional[Dict[str, Union[StrictStr, StrictInt, StrictFloat]]]

    class Config:
        arbitrary_types_allowed = True

    @validator('weight', 'order')
    def validate_weight_and_order(cls, val):
        assert val > 0

        return val
